from pathlib import Path
import os
import tarfile
from glob import iglob
import random
from torchvision.datasets import utils

from . import ClassificationDataset


def extract(src, dst=None):
  dst = dst or os.path.dirname(src)
  with tarfile.open(src, 'r:gz') as tar:
      tar.extractall(path=dst)


class Imagenette(ClassificationDataset):
  """
  https://github.com/fastai/imagenette
  """
  urls = {
    160: 'https://s3.amazonaws.com/fast-ai-imageclas/imagenette-160.tgz',
    320: 'https://s3.amazonaws.com/fast-ai-imageclas/imagenette-320.tgz',
    None: 'https://s3.amazonaws.com/fast-ai-imageclas/imagenette.tgz'
  }

  @classmethod
  def get_dirname(cls, url):
    return url.split('/')[-1][:-4]

  @classmethod
  def get_filename(cls, url):
    return url.split('/')[-1]

  def __init__(self, root='./datasets/', size=None, split='train', download=True, shuffle=True):
    if size not in self.urls:
      raise ValueError(
        f"Unsupported size '{size}', "
        f"must be one of '{', '.join(self.urls.keys())}'")
    self.size = size
    self.root = Path(root).expanduser()
    self.size_root = self.root / self.get_dirname(self.urls[self.size])

    self.split = split

    if not self.size_root.exists():
      if not download:
        raise ValueError(
          f'Could not find dataset at provided root ({self.size_root}) and download=False'
        )
      self.download()

    paths = list(iglob(f'{self.size_root}/{self.split}/**/*.*', recursive=True))
    if shuffle:
      random.shuffle(paths)
    self.inputs, self.targets = list(zip(*(
      (p, p.split('/')[-2]) for p in paths
    )))

    super(Imagenette, self).__init__(classes=sorted(set(self.targets)))

  @property
  def paths(self):
    return self.inputs

  def download(self):
    self.size_root.mkdir(parents=True, exist_ok=True)
    url = self.urls[self.size]
    root = self.root
    filename = self.get_filename(url)
    utils.download_url(url, root, filename)
    print('extracting file', root / filename)
    extract(root / filename)


class Imagenette160(Imagenette):
  size = 160
  def __init__(self, **kwargs):
    if 'size' in kwargs and kwargs['size'] != self.__class__.size:
      raise ValueError('size is not a valid argument')
    super(Imagenette160, self).__init__(
      size=self.__class__.size,
      **kwargs
    )

class Imagenette320(Imagenette):
  size = 320
  def __init__(self, **kwargs):
    if 'size' in kwargs and kwargs['size'] != self.__class__.size:
      raise ValueError('size is not a valid argument')
    super(Imagenette320, self).__init__(
      size=self.__class__.size,
      **kwargs
    )

class Imagewoof(Imagenette):
  urls = {
    160: 'https://s3.amazonaws.com/fast-ai-imageclas/imagewoof-160.tgz',
    320: 'https://s3.amazonaws.com/fast-ai-imageclas/imagewoof-320.tgz',
    None: 'https://s3.amazonaws.com/fast-ai-imageclas/imagewoof.tgz'
  }


class Imagewoof160(Imagenette):
  size = 160
  def __init__(self, **kwargs):
    if 'size' in kwargs and kwargs['size'] != self.__class__.size:
      raise ValueError('size is not a valid argument')
    super(Imagewoof160, self).__init__(
      size=self.__class__.size,
      **kwargs
    )


class Imagewoof320(Imagenette):
  size = 320
  def __init__(self, **kwargs):
    if 'size' in kwargs and kwargs['size'] != self.__class__.size:
      raise ValueError('size is not a valid argument')
    super(Imagewoof320, self).__init__(
      size=self.__class__.size,
      **kwargs
    )


from torch import nn

from ...data import Classes
from typing import Callable



class CNN(nn.Module):
  input_shape = (None, None, None, None)
  backbone: Callable

  def initialize(self):
    pass


class RecognitionModel(nn.Module):
  input_shape = (None, None, None, None)
  classes: Classes


  backbone: Callable
  # "Global" pooling layer that converts features into an embedding
  pool_features: Callable

  classifier: Callable


  def get_features(self, inputs):
    raise NotImplementedError()

  def get_embeddings(self, inputs):
    raise NotImplementedError()

  def get_logits(self, inputs):
    raise NotImplementedError()

  def freeze(self):
    pass

  def unfreeze(self):
    pass

  def freeze_backbone(self):
    pass

  def replace_classifier(self, linear=None, num_classes=None, init=None):
    raise NotImplementedError()


from torch.autograd.profiler import profile
from typing import Union, Tuple, Callable
import torch
from .timer import Timer

# TODO:
# - https://github.com/pytorch/pytorch/issues/3749#issuecomment-374006211
# - https://pytorch.org/docs/stable/bottleneck.html

def param_count(model):
  return sum(p.numel() for p in model.parameters())


def profile_module(
    module,
    input,
    warmup=10,
    iterations=20,
    sync=True,
    timer=None,
    task_name='iteration',
    jit=False
):
  if jit:
    import torch.jit
    module = torch.jit.trace(module, input)

  for n in range(warmup):
    module(input)

  t = timer or Timer()
  tasks = []
  for n in range(iterations):
    with t.task(task_name, sync=sync) as task:
      module(input)
      tasks.append(task)

  return tasks



from torch import nn


class DepthwiseConv2d(nn.Conv2d):
  def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1,
                 bias=True, padding_mode='zeros'):
    if out_channels % in_channels:
      raise ValueError('out_channels must be a multiple of in_channels')
    super(DepthwiseConv2d, self).__init__(
      in_channels=in_channels,
      out_channels=out_channels,
      kernel_size=kernel_size,
      stride=stride,
      padding=padding,
      dilation=dilation,
      bias=bias,
      padding_mode=padding_mode,
      groups=in_channels
    )


class DepthwiseSeparableConv2d(nn.Module):
  def __init__(self, in_channels, out_channels, kernel_size, stride=1,
                 padding=0, dilation=1,
                 bias=True, padding_mode='zeros'):
    super(DepthwiseSeparableConv2d, self).__init__()

    self.depthwise = DepthwiseConv2d(in_channels=in_channels,
      out_channels=out_channels,
      kernel_size=kernel_size,
      stride=stride,
      padding=padding,
      dilation=dilation,
      bias=bias,
      padding_mode=padding_mode)

    self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1)

  def forward(self, input):
    x = self.depthwise(input)
    return self.pointwise(x)


from typing import Union, Iterable, Any
from glob import iglob
import os

import torch
from torch.utils import data

from .wrappers import LookupCache, DatasetWrapper, TransformDataset, IncludeIndex

from ..classes import Classes


class Dataset(data.Dataset):
  def state_dict(self):
    return {
      'name': self.__class__.__name__
    }


class SupervisedDataset(Dataset):
  def __init__(self):
    if not hasattr(self, 'inputs'):
      self.inputs = None
    if not hasattr(self, 'targets'):
      self.targets = None

  def __getitem__(self, idx):
    return self.inputs[idx], self.targets[idx]

  def __len__(self):
    return len(self.inputs)


class ClassificationDataset(SupervisedDataset):
  def __init__(self, classes: Union[Classes, Iterable[Any]]):
    super(ClassificationDataset, self).__init__()
    self.classes = classes if isinstance(classes, Classes) else Classes(classes)


class GlobDataset(Dataset):
  def __init__(self, pattern='**/*.*', limit=None):
    paths = []
    for n, p in enumerate(iglob(pattern, recursive=True)):
      if os.path.getsize(p) < 4000:
        continue
      paths.append(p)
      if limit and n >= limit:
        break
    self.paths = paths

  def __len__(self):
    return len(self.paths)

  def __getitem__(self, idx):
    return (self.paths[idx], 0)


class InputsTargetsDataset(Dataset):
  def __init__(self, inputs, targets, transform=None):
    self.inputs = inputs
    self.targets = targets
    self.transform = transform
    self.classes = Classes(sorted(set(x for t in self.targets for x in t)))

  def __len__(self):
    return len(self.inputs)

  def __getitem__(self, idx):
    x, y = self.inputs[idx], self.targets[idx]
    if self.transform:
      x = self.transform(x)
    return x, y


class TinyDigits(data.TensorDataset):
  """
  Dataset of 8x8 digits, best used for testing
  """

  def __init__(self, num_classes=10):
    from sklearn.datasets import load_digits
    digits = load_digits(num_classes)
    super().__init__(
      torch.from_numpy(digits.images).unsqueeze(1).float(),
      torch.Tensor(digits.target).long()
    )





from yann.callbacks.base import Callback
from torch.utils.tensorboard import SummaryWriter


class Tensorboard(Callback):
  def __init__(self, root=None):
    self.root = root
    self.writer = None
    self.trainer = None

  def open(self):
    self.writer = SummaryWriter(self.root)

  def close(self):
    self.writer.close()

  def on_train_start(self, trainer=None):
    self.trainer = trainer
    if not self.root:
      self.root = self.trainer.root / 'tensorboard'
    self.open()

  def on_train_end(self, trainer=None):
    self.close()

  def help(self):
    print(
      f'''
      # Viewing in Jupyter:
      %load_ext tensorboad
      %tensorboard --logdir={self.root}
      '''
    ) 


import torch



def device_info(d=None):
  d = d or torch.cuda.current_device()
  return {
    'name': torch.cuda.get_device_name(d),
    'capability': torch.cuda.get_device_capability(d),
    'allocated_memory': torch.cuda.memory_allocated(d),
    'cached_memory': torch.cuda.memory_cached(d),
  }


sync = torch.cuda.synchronize


from torch import nn


def replace_last_linear(model, num_classes, layer_name='last_linear'):
  setattr(
    model,
    layer_name,
    nn.Linear(
      getattr(model, layer_name).in_features,
      num_classes
    )
  )


from yann.modules.loss import binary_focal_loss
import torch
from torch.nn.functional import binary_cross_entropy_with_logits

from yann.testing import check_tensor

T = torch.tensor


def test_focal_loss():
  logits = torch.Tensor([
      [-10, 2.3],
      [-10, 5.5]
    ])

  targets = torch.Tensor([
      [1.0, 0.0],
      [.0, 1.0]
    ])

  assert torch.allclose(
    binary_focal_loss(
      logits, targets, reduction='none', gamma=0, alpha=None),
    binary_cross_entropy_with_logits(
      logits, targets, reduction='none')
  ), "focal loss with gamma == 0 and no alpha should be same as binary cross entropy loss"

  assert torch.allclose(
    binary_focal_loss(
      logits, targets, reduction='mean', gamma=0, alpha=None),
    binary_cross_entropy_with_logits(
      logits, targets, reduction='mean')
  ), "focal loss with gamma == 0 and no alpha should be same as binary cross entropy loss"

  assert not torch.allclose(
    binary_focal_loss(
      logits, targets, reduction='none', gamma=.4, alpha=None),
    binary_cross_entropy_with_logits(
      logits, targets, reduction='none')
  ), "focal loss with gamma > 0 should not be same as binary cross entropy loss"


  assert binary_focal_loss(
    T([[100.0]]),
    T([[1.0]]),
    gamma=4
  ) == binary_focal_loss(
    T([[-100.0]]),
    T([[0.0]]),
    gamma=4
  ), "loss should be symmetrical"

  assert binary_focal_loss(
    T([[100.0]]),
    T([[0.0]])
  ) == binary_focal_loss(
    T([[-100.0]]),
    T([[1.0]])
  )

  assert binary_focal_loss(
    T([[2.0]]),
    T([[1.0]]),
    gamma=1
  ) > binary_focal_loss(
    T([[2.0]]),
    T([[1.0]]),
    gamma=2
  ), "larger gamma should reduce loss for well classified examples"

  assert binary_focal_loss(
    T([[2.0]]),
    T([[1.0]]),
    gamma=1
  ) < binary_cross_entropy_with_logits(
    T([[2.0]]),
    T([[1.0]]),
  ), "focal loss should penalize well classified examples less than binary cross entropy"

  for g in range(0, 10):
    for logit in range(-1_000, 1_000, 100):
      check_tensor(
        binary_focal_loss(
          T([[float(logit)]]),
          T([[1.0]]),
          gamma=g
        ),
        anomalies=True,
        gte=0,
      )
  check_tensor(
    binary_focal_loss(
      T([[-.3]]),
      T([[.2]]),
      gamma=2
    ),
    anomalies=True,
    gte=0,
  )

def test_focal_loss_gradients():
  logits = T([[0.0], [0.0]], requires_grad=True)
  targets = T([[0.0], [0.0]])

  loss = binary_focal_loss(logits, targets)
  loss.backward()
  check_tensor(logits.grad, anomalies=True)


import math


def get_tf_same_padding(
    tensor,
    kernel_size,
    stride=1,
    dilation=1
):
  if isinstance(kernel_size, int):
    return tuple(
      *tf_same_pad(tensor.shape[2], kernel_size, stride=stride, dilation=dilation),
      *tf_same_pad(tensor.shape[3], kernel_size, stride=stride, dilation=dilation)
    )
  else:
    return tuple(
      *tf_same_pad(tensor.shape[2], kernel_size[0], stride=stride, dilation=dilation),
      *tf_same_pad(tensor.shape[3], kernel_size[1], stride=stride, dilation=dilation)
    )

def tf_same_pad(size, kernel_size, stride=1, dilation=1):
  pad = max(
    0,
    (math.ceil(size / stride) - 1) * stride
      + (kernel_size - 1) * dilation
      + 1 - size
  )
  left_pad = pad // 2
  return left_pad, pad - left_pad


def get_same_padding(kernel_size, stride=1, dilation=1):
  return ((stride - 1) + dilation * (kernel_size - 1)) // 2


import torch

def check_tensor(
    t: torch.Tensor,
    device=None,
    dtype=None,
    requires_grad=None,
    contiguous=None,
    pinned=None,
    shape=None,
    anomalies=True,
    equal=None,
    close=None,
    lt=None,
    gt=None,
    lte=None,
    gte=None,
    none=False
):
  if not none:
    assert t is not None
  if device:
    if isinstance(device, str):
      device = torch.device(device)
    assert t.device == device
  if dtype:
    assert t.dtype == dtype
  if requires_grad is not None:
    assert t.requires_grad == requires_grad
  if contiguous is not None:
    assert t.is_contiguous() == contiguous
  if pinned is not None:
    assert t.is_pinned() == pinned
  if shape:
    assert t.shape == torch.Size(shape)
  if anomalies:
    assert not torch.isnan(t).any() and not torch.isinf(t).any()
  if equal is not None:
    if not torch.is_tensor(equal):
      equal = torch.Tensor(equal, dtype=t.dtype, device=t.device)
    assert (t == equal).all()
  if close is not None:
    if not torch.is_tensor(close):
      close = torch.Tensor(close, dtype=t.dtype, device=t.device)
    assert torch.allclose(t, close)
  if lt is not None:
    assert (t < lt).all()
  if lte is not None:
      assert (t <= lte).all(), lte
  if gt is not None:
    assert (t > gt).all(), t > gte
  if gte is not None:
    assert (t >= gte).all(), t >= gte


# def profile(name=None, sync=True, max=None, track=False, log=True):
#   """
#   track execution speed of code block
#   Args:
#     sync: cuda synchronize or not
#     max: fail if takes longer than x amount of time
#     name:
#     track:
#
#   Returns:
#
#   """
#   pass
#
#
# def track_value():
#   pass

#
# with profile('forward pass', max=.3):
#   pass


# def assert_finite(x):
#   pass
#
#
# def check_stability():
#   pass
#
#
# def check_numerical_issues(function, shapes, types):
#   pass




from matplotlib import pylab as plt
import numpy as np
import datetime
import pathlib
from sklearn.metrics import roc_curve, auc, confusion_matrix
import matplotlib.dates as mdates
from matplotlib.collections import PolyCollection
from collections import OrderedDict

import itertools

from .. import to_numpy
from ..metrics import moving_average

def plot_line(
    y,
    x=None,
    figsize=None,
    window=1,
    xlim=None,
    ylim=None,
    line_width=2,
    stroke='-',
    title=None,
    xlabel=None,
    ylabel=None,
    xscale=None,
    yscale=None,
    show=True,
    save=False,
    legend=True,
    name=None,
    grid=True
):


  fig = figsize and plt.figure(figsize=figsize)

  if xlim:
    plt.xlim(xlim)
  if ylim:
    plt.ylim(ylim)

  if xlabel:
    plt.xlabel(xlabel)
  if ylabel:
    plt.ylabel(ylabel)
  if title:
    plt.title(title)

  if yscale:
    plt.yscale(yscale)

  if xscale:
    plt.xscale(xscale)

  if window != 1:
    y = moving_average(y, window=window)

  if x is None:
    x = [n * window for n in range(len(y))]
  elif len(x) != len(y):
    x = moving_average(x, window=window)

  plt.plot(x, y, stroke, lw=line_width, label=name and str(name))
  if grid:
    plt.grid()

  if legend:
    plt.legend(loc='best')

  if save and show:
    raise ValueError("Can't show and save at the same time")
  if show:
    plt.show()
  if save:
    plt.savefig(save if isinstance(save, (str, pathlib.Path)) else
                f"{title or ylabel or datetime.datetime.utcnow().strftime('%y-%m-%dT%H%M%S')}.jpg")
    plt.gcf().clear()
    if fig:
      plt.close(fig)


def plot_pred_scores(
    preds,
    targets,
    classes=None,
    logscale=True,
    figsize=(12, 6),
    show=True,
    save=False,
    title=None):
  import seaborn as sns

  preds, targets = to_numpy(preds), to_numpy(targets)

  if title:
    plt.title(title)


  if classes:
    classes = classes.items() if isinstance(classes, dict) else \
      ((c, n) for n, c in enumerate(classes))
  else:
    classes = ((n, n) for n in range(preds.shape[1]))


  for cls, idx in classes:
    f, ax = plt.subplots(figsize=figsize)
    if logscale:
      ax.set(yscale="log")

    if len(targets.shape) == 1:
      sns.distplot(preds[targets != idx, idx], bins=50, kde=False,
                   rug=False, hist_kws={"range": [0, 1]}, ax=ax, color='red',
                   label='Negative')
      sns.distplot(preds[targets == idx, idx], bins=50, kde=False,
                   rug=False, hist_kws={"range": [0, 1]}, ax=ax, color='blue',
                   label='Positive')
    else:
      sns.distplot(preds[targets[:, idx] == 0, idx], bins=50, kde=False,
                   rug=False, hist_kws={"range": [0,1]}, ax=ax, color='red', label='Negative')
      sns.distplot(preds[targets[:, idx] == 1, idx], bins=50, kde=False,
                   rug=False, hist_kws={"range": [0,1]}, ax=ax, color='blue', label='Positive')
    ax.set_title(cls)
    plt.xlabel('Score')
    plt.ylabel('Sample Count')
    plt.legend(loc='best')

  if show:
    plt.show()
  if save:
    plt.savefig(save if isinstance(save, (str, pathlib.Path)) else
                  f"{title or datetime.datetime.utcnow().strftime('%y-%m-%dT%H%M%S')}.jpg")
    plt.gcf().clear()


def plot_rocs(
    preds,
    targets,
    classes=None,
    figsize=(12,12),
    show=True,
    save=False,
    title=None):

  preds, targets = to_numpy(preds), to_numpy(targets)

  fig = plt.figure(figsize=figsize)
  if title:
    plt.title(title)

  if classes:
    classes = classes.items() if isinstance(classes, dict) else \
      ((c, n) for n, c in enumerate(classes))
  else:
    classes = ((n, n) for n in range(preds.shape[1]))

  plt.xlim([0.0, 1.02])
  plt.ylim([0.0, 1.02])
  plt.xlabel('False Positive Rate')
  plt.ylabel('True Positive Rate')
  plt.plot([0, 1], [0, 1], 'k--', lw=2)


  vals = {}
  for cls, idx in classes:
    if len(targets.shape) == 1:
      fpr, tpr, thresholds = roc_curve(targets, preds[:, idx])
    else:
      fpr, tpr, thresholds = roc_curve(targets[:, idx], preds[:, idx])
    area = auc(fpr, tpr)
    plt.plot(fpr, tpr, label=f'{cls} (AUC = %0.3f)' % area)

    vals[cls] = (area, fpr, tpr, thresholds)
  plt.legend(loc='best')


  if show:
    plt.show()
  if save:
    plt.savefig(save if isinstance(save, (str, pathlib.Path)) else
                f"{title or datetime.datetime.utcnow().strftime('%y-%m-%dT%H%M%S')}.jpg")
    plt.gcf().clear()

  return vals


def plot_cooccurrences(counts, classes, figsize=(14, 11)):
  import seaborn as sns
  import pandas as pd

  mask = np.ones_like(counts)
  mask[np.tril_indices_from(mask)] = False

  df_cm = pd.DataFrame(
    counts,
    index=list(classes),
    columns=list(classes))
  plt.figure(figsize=figsize)
  plot = sns.heatmap(
    df_cm,
    robust=True,
    annot=True,
    fmt="d",
    cmap="YlGnBu",
    mask=mask)
  plot.set_title('Class Co-occurrence')


def sorted_indices(matrix, desc=False):
  inds = np.dstack(np.unravel_index(np.argsort(matrix.ravel()), matrix.shape))
  return (np.fliplr(inds) if desc else inds)[0]


def truncate_confusion_matrix(matrix, thresh=.2, top=None, symmetric=True):
  inds = sorted_indices(matrix, desc=True)
  inds = [(r, c) for r, c in inds if r != c]

  if top:
    inds = inds[:top]

  if thresh:
    inds = [(r, c) for r, c in inds if matrix[r, c] >= thresh]

  if symmetric:
    rows = cols = sorted(set(y for x in inds for y in x))
  else:
    rows, cols = zip(*inds)
    rows, cols = sorted(set(rows)), sorted(set(cols))

  m = matrix[rows, :][:, cols]
  return m, rows, cols


def plot_confusion_matrix(
    preds, targets, classes, figsize=(16, 16),
    thresh=None, top=None, normalize=False, symmetric=True):
  preds, targets = to_numpy(preds), to_numpy(targets)

  cm = confusion_matrix(targets, preds)

  if normalize:
    cm = cm.astype('float') / cm.sum(axis=1)[:, np.newaxis]

  if thresh or top:
    cm, rows, cols = truncate_confusion_matrix(
      cm, thresh=thresh, top=top, symmetric=symmetric)
  else:
    rows, cols = list(range(len(classes))), list(range(len(classes)))

  if not len(cm):
    return cm

  fig = plt.figure(figsize=figsize)
  plt.imshow(cm, interpolation='nearest', cmap=plt.cm.Blues)
  plt.title('Confusion matrix')
  plt.colorbar()
  plt.xticks(np.arange(len(cols)), [classes[i] for i in cols], rotation=75)
  plt.yticks(np.arange(len(rows)), [classes[i] for i in rows])

  fmt = '.2f' if normalize else 'd'
  th = cm.max() / 2.
  for i, j in itertools.product(range(cm.shape[0]), range(cm.shape[1])):
    plt.text(j, i, format(cm[i, j], fmt),
             horizontalalignment="center",
             color="white" if cm[i, j] > th else "black")

  plt.tight_layout()
  plt.ylabel('True label')
  plt.xlabel('Predicted label')
  plt.show()

  return cm


def plot_timeline(tasks, figsize=None):
  inds = OrderedDict((c, n) for n, c in enumerate(set(t.name for t in tasks)))
  color_map = {c: f"C{n}" for c, n in inds.items()}

  verts = []
  colors = []
  for task in tasks:
    v = [(mdates.date2num(task.start_time), inds[task.name] - .4),
         (mdates.date2num(task.start_time), inds[task.name] + .4),
         (mdates.date2num(task.end_time), inds[task.name] + .4),
         (mdates.date2num(task.end_time), inds[task.name] - .4),
         (mdates.date2num(task.start_time), inds[task.name] - .4)]
    verts.append(v)
    colors.append(color_map[task.name])

  bars = PolyCollection(
    verts,
    facecolors=colors,
  )

  fig, ax = plt.subplots(figsize=figsize)
  ax.add_collection(bars)
  ax.autoscale()
  loc = mdates.SecondLocator(list(range(0, 60, 10)))
  ax.xaxis.set_major_locator(loc)
  ax.xaxis.set_major_formatter(mdates.AutoDateFormatter(loc))

  ax.set_yticks(list(inds.values()))
  ax.set_yticklabels(list(inds.keys()))
  plt.show()


from torch.nn import functional as F



def mac(batch):
  """MAC Pooling"""
  return F.adaptive_max_pool2d(batch, (1, 1))

def spoc(batch):
  """SPoC Pooling"""
  return F.adaptive_avg_pool2d(batch, (1, 1))


def generalized_mean(batch, p=3, eps=1e-8):
  """Generalized Mean Pooling (GeM)
  p=1 === spoc
  larger p ==> mac

  larger p leads to more localized (max) features
  """
  return F.adaptive_avg_pool2d(
    batch.clamp(min=eps) ** p,
    (1, 1)
  ) ** (1 / p)

gem = generalized_mean


def blur_pool():
  """
  https://richzhang.github.io/antialiased-cnns/
  Returns:

  """
  raise NotImplementedError()


class StackedPooling:
  pass


import torch
from torch.utils.data.dataloader import default_collate
from .utils import pad_to_largest


class PadCollate:
  def __init__(self, pad_inputs=True, pad_targets=False, value=0):
    self.pad_inputs = pad_inputs
    self.pad_targets = pad_targets
    self.value = value

  def __call__(self, batch):
    inputs, targets = zip(*batch)

    if self.pad_inputs:
      inputs = pad_to_largest(list(inputs), self.value)
    if self.pad_targets:
      targets = pad_to_largest(list(targets), self.value)

    return torch.stack(inputs), torch.stack(targets)



class FilterCollate:
  def __init__(self, filter=None, collate=default_collate, value=None):
    self.filter = filter or (
      lambda items: [it for it in items if it is not value]
    )
    self.collate = collate

  def __call__(self, batch):
    filtered = self.filter(batch)
    return self.collate(filtered)


filter_collate = FilterCollate()





from collections import OrderedDict

from .tensor import weighted_sum


def average_state_dicts(dicts, weights=None):
  if not weights:
    weights = (1 / len(dicts),) * len(dicts)
  averaged = OrderedDict()
  for k in dicts[0]:
    averaged[k] = weighted_sum([d[k] for d in dicts], weights)
  return averaged



from torch import nn
import torch

from .utils import get_same_padding

class MixConv(nn.Module):
  convs: nn.ModuleList
  """
  
  From:
    MixConv: Mixed Depthwise Convolutional Kernels
    https://arxiv.org/abs/1907.09595
  
  
  NOTE: with MKLDNN and padding == 3 and depthwise=False there's a runtime error that crashes the interpreter
  so kernel_size = 7 might lead to crashes on CPUs (seeing this on a Mac) 
  https://github.com/pytorch/pytorch/issues/20583
  """
  def __init__(
      self,
      in_channels,
      out_channels,
      kernel_size=None,
      depthwise=True
  ):
    super(MixConv, self).__init__()

    if isinstance(kernel_size, int):
      self.kernel_sizes = [kernel_size]
    elif isinstance(kernel_size, (tuple, list)):
      self.kernel_sizes = list(kernel_size)
    elif kernel_size is None:
      if not isinstance(in_channels, (list, tuple)):
        raise ValueError('kernel_size must be provided if in_channels is not an iterable')
      self.kernel_sizes = [3 + 2 * n for n in range(len(in_channels))]

    if isinstance(in_channels, (list, tuple)):
      self.input_channel_counts = in_channels
    else:
      self.input_channel_counts = self.split_groups(in_channels, len(self.kernel_sizes))
    if isinstance(out_channels, (list, tuple)):
      self.output_channel_counts = out_channels
    else:
      self.output_channel_counts = self.split_groups(out_channels, len(self.kernel_sizes))

    if len(self.input_channel_counts) != len(self.output_channel_counts):
      raise ValueError(
        f'in_channels and out_channels should have same number of groups,'
        f' but got {len(self.input_channel_counts)} and {len(self.output_channel_counts)}'
      )

    self.convs = nn.ModuleList([
      nn.Conv2d(
        in_channels=ic,
        out_channels=oc,
        kernel_size=ks,
        groups=min(ic, oc) if depthwise else 1,
        padding=get_same_padding(ks)
      )
      for ic, oc, ks
      in zip(self.input_channel_counts, self.output_channel_counts, self.kernel_sizes)
    ])

  def __repr__(self):
    return f"MixConv({self.input_channel_counts}, {self.output_channel_counts}, kernel_sizes={self.kernel_sizes}, convs={self.convs})"

  def split_groups(self, num_channels, num_groups):
    channel_counts = [num_channels // num_groups] * num_groups
    channel_counts[0] += num_channels - sum(channel_counts)
    return channel_counts

  def forward(self, input):
    if len(self.convs) == 1:
      return self.convs[0](input)
    parts = torch.split(input, self.input_channel_counts, 1)
    outputs = [conv(part) for conv, part in zip(self.convs, parts)]
    return torch.cat(outputs, 1)





import base64
import os

import io
import numpy as np
import pathlib
import torch
import random
from PIL import Image
from torchvision import transforms as tvt
from torchvision.transforms.functional import to_pil_image
from torchvision import transforms

from ..utils import truthy


class Transformer:
  def __init__(self, load=None, transform=None, to_tensor=None):
    self.load = load
    self.transform = transform
    self.to_tensor = to_tensor

  def __call__(self, x):
    x = self.load(x)
    x = self.transform(x)
    return self.to_tensor(x)

  def prep(self, x):
    return self.transform(self.load(x))

  def __repr__(self):
    return (
      f"{self.__class__.__name__}(\n"
      f"  load={str(self.load)}\n"
      f"  transform={str(self.transform)}\n"
      f"  to_tensor={str(self.to_tensor)}\n"
      ")")


class ImageTransformer(Transformer):
  def __init__(
      self,
      resize=None,
      rotate=None,
      crop=None,
      warp=None,
      mirror=None,
      mean=None,
      std=None,
      color_jitter=None,
      interpolation=None,
      color_space=None,
      load=None,
      transform=None,
      to_tensor=None
  ):
    interpolation = interpolation or Image.ANTIALIAS
    self.resize = resize and tvt.Resize(resize, interpolation=interpolation)
    self.rotate = rotate and tvt.RandomRotation(rotate)
    self.crop = crop and (
      tvt.RandomResizedCrop(crop, interpolation=interpolation)
      if warp
      else tvt.CenterCrop(crop)
    )
    self.mirror = mirror and tvt.RandomHorizontalFlip(
      .5 if mirror is True else mirror)

    if color_jitter is True:
      color_jitter = (.4, .2, .1, .05)
    self.color_jitter = color_jitter and tvt.ColorJitter(*color_jitter)

    self.normalize = (mean or std) and tvt.Normalize(mean=mean, std=std)

    super().__init__(
      load=load or GetImage(color_space),
      transform=tvt.Compose(truthy([
        self.resize,
        transform,
        self.rotate,
        self.crop,
        self.mirror,
        self.color_jitter,
      ])),
      to_tensor=to_tensor or tvt.Compose(truthy([
        tvt.ToTensor(),
        self.normalize
      ]))
    )


class GetImage:
  def __init__(self, space=None):
    self.color_space = space

  def __call__(self, x):
    return get_image(x, self.color_space)


def get_image(x, space=None) -> Image.Image:
  if isinstance(x, Image.Image):
    return x.convert(space) if space else x

  if isinstance(x, np.ndarray):
    img = Image.fromarray(x)
    return img.convert(space) if space else img

  if isinstance(x, torch.Tensor):
    img = to_pil_image(x)
    return img.convert(space) if space else img

  if isinstance(x, (str, pathlib.Path)) and os.path.exists(x):
    img = Image.open(x)
    return img.convert(space) if space else img

  if isinstance(x, str):
    if x.startswith('http') or x.startswith('www.'):
      import requests
      x = requests.get(x).content
    elif x.startswith('data') and 'base64,' in x:
      # data header for base64 encoded
      x = x.split('base64,')[1]
      x = base64.b64decode(x)
    elif len(x) > 1024:
      # assume no header base 64 image
      try:
        x = base64.b64decode(x)
      except:
        pass

  if hasattr(x, 'read'):
    img = Image.open(io.BytesIO(x.read()))
    return img.convert(space) if space else img

  if isinstance(x, bytes):
    img = Image.open(io.BytesIO(x))
    return img.convert(space) if space else img




def mixup(inputs, targets, alpha=1):
  """
  Args:
    inputs: batch of inputs
    targets: hot encoded targets
  Returns:
    mixed up (inputs, targets)

  """
  shuffled_indices = torch.randperm(inputs.shape[0], device=inputs.device)
  fraction = np.random.beta(alpha, alpha)
  return (
    fraction * inputs + (1 - fraction) * inputs[shuffled_indices],
    fraction * targets + (1 - fraction) * targets[shuffled_indices]
  )



def cutout(img, percent=.3, value=0):
  pil_img = False
  if isinstance(img, Image.Image):
    img = np.array(img)
    pil_img = True
  height, width = img.shape[:2]

  mask_height = round(height * percent)
  mask_width = round(width * percent)

  start_h = random.randint(0, (height - mask_height))
  start_w = random.randint(0, (width - mask_width))

  img[start_h:start_h + mask_height, start_w:start_w + mask_width] = value
  return Image.fromarray(img) if pil_img else img



def get_imagenet_transformers(size=224, resize=256, fixres=False):
  train_transform = Transformer(
    load=GetImage('RGB'),
    transform=transforms.Compose([
#       transforms.Resize(resize, interpolation=Image.ANTIALIAS),
      transforms.RandomResizedCrop(
          size,
#           scale=(.4, 1),
#           interpolation=Image.ANTIALIAS
      ),
      transforms.ColorJitter(
          .3, .3, .3
#           brightness=.4, contrast=.2, saturation=.1, hue=.05
      ),
      transforms.RandomHorizontalFlip(),
    ]),
    to_tensor=transforms.Compose([
      transforms.ToTensor(),
      transforms.Normalize(
          mean=[0.485, 0.456, 0.406],
          std=[0.229, 0.224, 0.225])
    ])
  )

  test_transform = Transformer(
    load=train_transform.load,
    transform=transforms.Compose([
      transforms.Resize(size, interpolation=Image.ANTIALIAS),
      transforms.CenterCrop(size)
    ]),
    to_tensor=train_transform.to_tensor
  )

  return train_transform, test_transform



import urllib.request
from concurrent import futures
import pathlib
from pathlib import Path
from urllib.parse import urlparse
import os

from ...utils import progress


class CachedExecutor:
  def __init__(self, workers=8):
    self._executor = futures.ThreadPoolExecutor(max_workers=workers)

    self.pending = {}  # key => future
    self.results = {} # key => local path
    self.errors = {}  # key => error

    self.error_callbacks = []
    self.success_callbacks = []
    self.cancel_callbacks = []

  def on_error(self, callback):
    self.error_callbacks.append(callback)

  def on_success(self, callback):
    self.success_callbacks.append(callback)

  def on_cancel(self, callback):
    self.cancel_callbacks.append(callback)

  def __del__(self):
    self.cancel()
    if self._executor:
      self._executor.shutdown()

  def cancel(self):
    for key, future in self.pending.items():
      future.cancel()

  def complete(self):
    self._executor.shutdown()

  def as_completed(self):
    yield from futures.as_completed(self.pending.values())

  def wait(self):
    return futures.wait(self.pending.values())

  def handle_done_future(self, future: futures.Future):
    self.pending.pop(future.key)

    if future.cancelled():
      for c in self.cancel_callbacks:
        c(future.key, future)
      return

    try:
      r = future.result()
      self.results[future.key] = r
    except Exception as e:
      self.errors[future.key] = e
      for c in self.error_callbacks:
        c(future.key, e, future)
      return

    for c in self.success_callbacks:
      c(future.key, r, future)

  def get_all(self, keys):
    fs = self._executor.map(self.get, keys)
    return futures.wait(fs)

  def stream(self, keys):
    fs = self._executor.map(self.get, keys)
    return futures.as_completed(fs)

  def execute(self, key, *args, **kwargs):
    raise NotImplementedError()

  def submit(self, key, *args, **kwargs):
    future = self._executor.submit(self.execute, key, *args, **kwargs)
    self.pending[key] = future
    future.key = key
    future.add_done_callback(self.handle_done_future)
    return future

  def enqueue(self, key, *args, **kwargs):
    if key in self.results or key in self.pending:
      return
    return self.submit(key,  *args, **kwargs)

  def get(self, key):
    if key in self.results:
      return self.results[key]
    if key in self.pending:
      return self.pending[key].result()

    return self.submit(key).result()


  def __getitem__(self, key):
    return self.get(key)


class Downloader(CachedExecutor):
  def __init__(self, local_root='./', workers=8):
    super(Downloader, self).__init__(workers=workers)
    self.local_root = local_root

  def execute(self, key, *args, **kwargs):
    uri = self.get_uri(key)
    dest = self.get_path(uri)
    return self.download(uri, dest)[0]

  def get_path(self, url):
    return None

  def get_uri(self, key):
    return key

  def download(self, uri, path):
    return download(uri, path, root=self.local_root)


def download(url, dest=None, skip_existing=True, nest=True, root='./'):
  """
  Returns: (local_path, headers), headers will be None if file exists
  """
  if not dest:
    root = os.path.abspath(root)
    dest = (urlparse(url).path if nest else os.path.basename(urlparse(url).path))
    dest = os.path.join(root, dest[1:] if dest[0] == '/' else dest)
  elif hasattr(dest, '__call__'):
    dest = dest(url)

  os.makedirs(os.path.dirname(dest), exist_ok=True)

  if skip_existing and pathlib.Path(dest).exists():
    return (dest, None)
  return urllib.request.urlretrieve(url, dest)


def download_urls(urls, dest=None, skip_existing=True, nest=True, root='./',
                  max_workers=12):
  results, errors = [], []
  with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
    queued_futures = {
      executor.submit(
        download,
        url,
        dest=dest,
        skip_existing=skip_existing,
        nest=nest,
        root=root
      ): url
      for url in urls
    }
    for f in progress(futures.as_completed(queued_futures)):
      try:
        results.append(f.result())
      except Exception as e:
        errors.append((queued_futures[f], e, f))
  return results, errors


import os
import shutil
import subprocess
from contextlib import suppress
from pathlib import Path

import torch

from .data import Classes
from .data.io import (
  load_pickle, load_json, save_json, save_pickle, tar_dir,
  untar
)


# TODO: add way to pass validation data to check model outputs when loaded again
def export(
    model=None,
    preprocess=None,
    postprocess=None,
    predict=None,
    classes=None,
    trace=False,
    state_dict=False,
    path=None,
    # validation=None,
    meta=None,
    tar=False,
    **kwargs
):
  os.makedirs(path, exist_ok=True)
  if os.listdir(path):
    raise ValueError(
      f'Failed to export because {path} already exists and is not empty')

  path = Path(path)

  if model:
    if trace is not False and trace is not None:
      from torch import jit
      traced = jit.trace(model, trace)
      traced.save(os.path.join(path, 'model.traced.th'))
    else:
      if not state_dict:
        torch.save(
          model,
          os.path.join(path, 'model.th')
        )
      else:
        torch.save(
          model.state_dict(),
          os.path.join(path, 'model.state_dict.th')
        )
  if preprocess:
    save_pickle(preprocess, path / 'preprocess.pkl')

  if postprocess:
    save_pickle(postprocess, path / 'postprocess.pkl')

  if meta:
    save_json(meta, path / 'meta.json')

  if classes:
    save_json(
      classes.state_dict() if isinstance(classes, Classes) else classes,
      path / 'classes.json'
    )

  if kwargs:
    save_pickle(kwargs, path / 'kwargs.pkl')

  if predict:
    save_pickle(predict, path / 'predict.pkl')

  subprocess.call(
    ['conda', 'env', 'export', '-f', os.path.join(path, 'env.yml')]
  )

  subprocess.call(
    ['pip', 'freeze'], stdout=open(os.path.join(path, 'requirements.txt'), 'w')
  )

  if tar:
    tar_dir(path)
    shutil.rmtree(str(path))


class Predictor:
  def __init__(self):
    self.model = None
    self.classes = None
    self.meta = {}
    self.kwargs = {}

    self.model_state_dict = None

    self.postprocess = None
    self.predict = None
    self.postprocess = None

  def predict(self):
    pass


def load(path, eval=True):
  path = Path(path)
  p = Predictor()

  # TODO: read from tarfile directly instead
  if str(path).endswith('.tar.gz'):
    untar(path)
    path = Path(str(path).rstrip('.tar.gz'))

  p.model = None
  if (path / 'model.th').exists():
    p.model = torch.load(str(path / 'model.th'))
  elif (path / 'model.traced.th').exists():
    from torch import jit
    p.model = jit.load(str(path / 'model.traced.th'))

  if p.model and eval:
    p.model.eval()

  with suppress(FileNotFoundError):
    p.model_state_dict = torch.load(str(path / 'model.state_dict.th'))

  with suppress(FileNotFoundError):
    p.classes = load_json(path / 'classes.json')

  with suppress(FileNotFoundError):
    p.preprocess = load_pickle(path / 'preprocess.pkl')

  with suppress(FileNotFoundError):
    p.postprocess = load_pickle(path / 'postprocess.pkl')

  with suppress(FileNotFoundError):
    p.predict = load_pickle(path / 'predict.pkl')

  with suppress(FileNotFoundError):
    p.meta = load_json(path / 'meta.json')

  with suppress(FileNotFoundError):
    p.kwargs = load_pickle(path / 'kwargs.pkl')

  return p



import logging
import os.path

from annoy import AnnoyIndex

from ..io import save_json, load_json
from .base import VectorIndex


class Annoy(VectorIndex):
  def __init__(self, path, dims=None, metric='angular', build_on_disk=True):
    self.path = path
    self.is_mutable = None
    self.is_built = None
    self.build_on_disk = build_on_disk
    self.metric = metric

    if os.path.isfile(self.path):
      logging.debug(f'Loading existing index: {self.path}')
      self.load_meta()
      assert self.dims == dims or not dims, \
        'Passed path to existing index but dims do not match'
      assert self.metric == metric or not metric, \
        'Passed path to existing index but metrics do not match'
      self.index = AnnoyIndex(self.dims, metric=self.metric)
    elif dims:
      logging.debug(
        f'Creating new index with {dims} dimensions and {self.metric} metric')
      self.dims = dims
      self.index = AnnoyIndex(self.dims, metric=self.metric)
      if build_on_disk:
        self.index.on_disk_build(self.path)
    else:
      logging.debug(f'Loading existing index: {self.path}')
      self.load_meta()
      self.index = AnnoyIndex(self.dims, metric=self.metric)

  @property
  def meta_path(self):
    return self.path + '.meta.json'

  @property
  def files(self):
    return [self.path, self.meta_path]

  def load_meta(self):
    self.__dict__.update(
      load_json(self.meta_path)
    )

  def save_meta(self):
    d = {**self.__dict__}
    d.pop('index')
    save_json(d, self.meta_path)

  def build(self, num_trees=10):
    logging.debug(f'staring to build index: {self.path}')
    self.index.build(num_trees)
    logging.debug(f'finished building index: {self.path}')
    self.is_mutable = False
    self.is_built = True
    self.save_meta()

  def save(self):
    self.index.save(self.path)
    self.is_mutable = False
    self.save_meta()

  def load(self, memory=False):
    self.index.load(self.path, prefault=memory)
    self.is_mutable = False

  def unload(self):
    self.index.unload()

  def __del__(self):
    self.unload()

  def __setitem__(self, idx, vector):
    self.index.add_item(idx, vector)

  def __getitem__(self, idx):
    return self.index.get_item_vector(idx)

  def __len__(self):
    return self.index.get_n_items()

  def add(self, vector):
    idx = len(self)
    self[idx] = vector
    return idx

  def add_bulk(self, vectors):
    start = len(self)
    for n, v in enumerate(vectors):
      self[start + n] = v
    return self

  def set_bulk(self, indices, vectors):
    for idx, vector in zip(indices, vectors):
      self[idx] = vector

  def search(self, vector, num=10, depth=None, distances=True):
    return self.index.get_nns_by_vector(vector, num, depth or -1, distances)

  def search_index(self, idx, num=10, depth=None, distances=True):
    return self.index.get_nns_by_item(idx, num, depth or -1, distances)

  def distance(self, i, j):
    return self.index.get_distance(i, j)



import torch

from .. import evalmode


class Predictor:
  def __call__(self, *args, **kwargs):
    x = self.load(*args, **kwargs)
    x = self.transform(x)
    x = self.predict(x)
    x = self.posprocess(x)

    return x

  def export(self):
    pass

  def load(self, samples):
    return samples

  def transform(self, inputs):
    return inputs

  def predict(self, inputs):
    return inputs

  def posprocess(self, inputs):
    return inputs


class Classifier(Predictor):
  def __init__(self, model, classes=None, preprocess=None, postprocess=None):
    self.model: torch.nn.Module = model
    self.preprocess = preprocess
    self.postprocess = postprocess

    self.classes = classes

  def __call__(self, *args, **kwargs):
    x = self.preprocess(*args, **kwargs)
    with evalmode(self.model):
      x = self.model(x)

    return self.classes.decode(x)



from .base import FunctionCallback, Callback
from .checkpoint import Checkpoint
from .eval import MulticlassEval
from .history import History, HistoryPlotter, HistoryWriter
from .logging import Logger
from .stop import StopOnNaN
from .timing import Timing


def get_callbacks(
    interactive=True,
    plot=True,
    write=True,
    log=True,
    checkpoint=True,
    time=False,
):
  return [
    x for x in (
      # History(),
      plot and HistoryPlotter(save=not interactive),
      write and HistoryWriter(),
      checkpoint and Checkpoint(),
      log and Logger(),
      time and Timing()
    ) if x]



import torch
import torch.nn.functional as F
from torch.nn.modules.loss import _Loss
from torch import nn

from ..data.classes import smooth as label_smoothing

def _reduce(x, reduce=True, reduction=None):
  if not reduce:
    return x
  if reduction == 'mean' or reduction == 'elementwise_mean':
    return x.mean()
  if reduction == 'sum':
    return x.sum()
  if reduction is None or reduction == 'none':
    return x
  raise ValueError(f'Unsupported reduction')

def soft_target_cross_entropy(
    inputs,
    targets,
    smooth=None,
    reduce=True,
    dim=1,
    reduction='mean'):
  """"like cross_entropy but using soft targets"""
  if smooth:
    targets = label_smoothing(targets, smooth)

  vals = torch.sum(-targets * F.log_softmax(inputs, dim=dim), dim=dim)
  return _reduce(vals, reduce=reduce, reduction=reduction)



class SoftTargetCrossEntropyLoss(_Loss):
  def __init__(self, smooth=None, reduce=True, dim=1, reduction='mean'):
    super().__init__(reduce=reduce, reduction=reduction)
    self.reduce = reduce
    self.reduction = reduction
    self.smooth = smooth
    self.dim = dim

  def forward(self, inputs, targets):
    return soft_target_cross_entropy(
      inputs,
      targets,
      smooth=self.smooth,
      reduce=self.reduce,
      dim=self.dim,
      reduction=self.reduction)



def binary_focal_loss(logits, targets, gamma=2, alpha=None, pos_weight=None, reduce=True, reduction='mean'):
  """
  Binary focal loss (with sigmoids)

  References:
    - https://arxiv.org/abs/1708.02002 - Focal Loss for Dense Object Detection
  Notes:
    - define:
        (pt = p if y = 1 else 1-p), where p is the model estimate for the class (y=1)
        modulating factor:  (1 − pt)^γ
        focusing parameter: γ  (gamma)
    - One notable property of cross entropy loss is that even examples that are
      easily classified (pt ≫ .5) incur a loss with non-trivial magnitude.
      When summed over a large number of easy examples, these small loss
      values can overwhelm the rare class.
    - When an example is misclassified and pt is small,
      the modulating factor is near 1 and the loss is unaffected.
      As pt → 1, the factor goes to 0 and the loss for well-classified
      examples is down-weighted.
    - The focusing parameter γ smoothly adjusts the rate at which easy
      examples are downweighted. When γ = 0, FL is equivalent to CE,
      and as γ is increased the effect of the modulating factor is
      likewise increased (we found γ = 2 to work best in our experiments)

  FL(pt) = −αt(1 − pt)^γ log(pt).

  Args:
    logits: logits (before activation is applied)
    targets: targets of same shape as logits
    gamma: "focusing parameter", gamma >= 0
    alpha: class balance parameters (class weights)

  Returns:

  """
  # TODO try this numerically stable version https://github.com/richardaecn/class-balanced-loss/issues/1

  probs = torch.sigmoid(logits)
  bce = F.binary_cross_entropy_with_logits(logits, targets, pos_weight=pos_weight, reduction='none')

  pt = targets * probs + (1 - targets) * (1 - probs)
  modulate = 1 if gamma is None else (1 - pt) ** gamma

  focal_loss = modulate * bce

  if alpha is not None:
    assert 0 <= alpha <= 1
    alpha_weights = targets * alpha + (1 - targets) * (1 - alpha)
    focal_loss *= alpha_weights

  return _reduce(focal_loss, reduce=reduce, reduction=reduction)


class BinaryFocalLoss(_Loss):
  def __init__(self, gamma=2, alpha=None, pos_weight=None, reduce=True, reduction='mean'):
    super(BinaryFocalLoss, self).__init__()

    self.gamma = gamma
    self.alpha = alpha
    self.pos_weight = pos_weight
    self.reduce = reduce
    self.reduction = reduction

  def forward(self, logits, targets):
    return binary_focal_loss(
      logits,
      targets,
      gamma=self.gamma,
      alpha=self.alpha,
      pos_weight=self.pos_weight,
      reduction=self.reduction
    )

class ClassWeighted(_Loss):
  def __init__(self, loss, weights=None, reduce=True, reduction='mean'):
    super(ClassWeighted, self).__init__(reduce=reduce, reduction=reduction)
    self.loss = loss(reduction='none') if issubclass(loss, _Loss) else loss
    self.register_buffer('weights', weights)
    self.reduce = reduce
    self.reduction = reduction

  def forward(self, outputs, targets):
    loss = self.loss(outputs, targets)
    if self.weights is not None:
      loss = loss * self.weights
    return _reduce(loss, reduce=self.reduce, reduction=self.reduction)


def contrastive_loss():
  pass


def triplet_loss():
  pass




def tempered_log(x, temperature=1):
  if temperature == 1: return torch.log(x)
  return (x ** (1 - temperature) - 1) / (1 - temperature)


def tempered_exp(x, temperature=1):
  if temperature == 1: return torch.exp(x)
  return torch.relu(1 + (1 - temperature) * x) ** (1 / (1 - temperature))


def bi_tempered_logistic_loss():
  """
  https://arxiv.org/pdf/1906.03361.pdf
  Returns:
  """
  pass


def bi_tempered_binary_logistic_loss():
  pass


import torch

def weighted_sum(tensors, weights):
  if len(tensors) < 2:
    raise ValueError('must pass at least 2 tensors')
  s = tensors[0] * weights[0]
  for t, w in zip(tensors[1:], weights[1:]):
    s.add_(w, t)
  return s


def one_hot(targets: torch.Tensor, num_classes=None, device=None, dtype=None, normalize=False):
  if torch.is_tensor(targets):
    if len(targets.shape) == 1:
      num = targets.shape[0]
      hot = torch.zeros(num, num_classes, device=device or targets.device, dtype=dtype)
      hot.scatter_(1, targets.unsqueeze(1), 1.0)
      return hot
    elif len(targets.shape) == 2:
      pass

    raise ValueError('only dim 1 tensors supported')



def show_hist(hist):
  chars = '_▁▂▃▄▅▆▇█'
  top = max(hist)
  step = (top / float(len(chars) - 1)) or 1
  return ''.join(chars[int(round(count / step))] for count in hist)


def describe(tensor: torch.Tensor, bins=10) -> str:
  try:
    stats = (
      f"μ={tensor.mean():.4f}  σ={tensor.std():.4f}\n"
    )
  except:
    stats = ''

  try:
    h = tensor.histc(bins=bins).int().tolist()
    hist = (
      f"hist: {h}\n"
      f"hist: {show_hist(h)}\n"
    )
  except:
    hist = ''
  return (

    f"{tuple(tensor.shape)} "
    f"{tensor.dtype} "
    f"{tensor.device}"
    f"{' grad' if tensor.requires_grad else ''} "
    f"({tensor.numel() * tensor.element_size() / (1e6):,.5f} MB)\n"
    f"{stats}"
    f"{hist}"
    f"min: {tensor.min():.4f}  max: {tensor.max():.4f}  sum: {tensor.sum():.4f}\n\n"
  )


from torch.utils.data import DataLoader, RandomSampler, SequentialSampler

from .datasets import TransformDataset


class TransformLoader(DataLoader):
  def __init__(self, dataset, transforms, **kwargs):
    super(TransformLoader, self).__init__(
      TransformDataset(dataset,  transforms),
      **kwargs
    )


from .base import Callback

import torch


class EarlyStopping(Callback):
  def __init__(self):
    pass


class StopOnNaN(Callback):
  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    if torch.isnan(loss).any() \
       or torch.isinf(loss).any():
      print('NaN or Inf detected, stopping training')
      trainer.stop()


import os
from math import cos, pi
import numpy as np

from ..callbacks.base import Callback
from .. import set_param
from ..metrics import exp_moving_avg
from ..viz import plot_line


def cosine_anneal(min_lr, max_lr, cur_step, num_steps):
  return min_lr + (max_lr - min_lr) * (1 + cos(cur_step / num_steps * pi)) / 2


class SGDR(Callback):
  def __init__(
      self,
      optimizer=None,
      max_lr=None,
      min_lr=0,
      cycle_len=10,
      cycle_mult=1,
      lr_mult=1,
      verbose=True,
      checkpoint=False
  ):
    self.optimizer = optimizer
    self.max_lr = max_lr
    self.min_lr = min_lr
    self.base_cycle_len = cycle_len
    self.cycle_mult = cycle_mult
    self.lr_mult = lr_mult

    self.cur_cycle_len = self.base_cycle_len
    self.cur_max_lr = self.max_lr
    self.cur_min_lr = self.min_lr
    self.cur_lr = self.max_lr
    self.cur_step = 0

    self.completed_cycles = 0
    self.save_checkpoints = checkpoint
    self.checkpoints = []

    self.verbose = verbose

    if self.cur_max_lr and self.optimizer:
      self.update_lr(self.cur_lr)

  def on_train_start(self, trainer=None):
    self.trainer = trainer

  def restart(self):
    if self.save_checkpoints:
      self.checkpoints.append(self.trainer.checkpoint(
        f'cycle-{self.completed_cycles}'
        f'-epochs-{self.trainer.num_epochs}'
        f'-steps-{self.trainer.num_steps}'))

    self.cur_cycle_len *= self.cycle_mult
    self.cur_max_lr *= self.lr_mult
    self.cur_min_lr *= self.lr_mult

    self.cur_lr = self.cur_max_lr
    self.cur_step = 0
    self.update_lr(self.cur_lr)

    if self.verbose:
      print('Restarting SGDR')

  def update_lr(self, lr):
    set_param(self.optimizer, 'lr', lr)

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    if self.cur_step >= self.cur_cycle_len:
      self.restart()
    else:
      new_lr = cosine_anneal(self.cur_min_lr, self.cur_max_lr, self.cur_step,
                             self.cur_cycle_len)
      self.update_lr(new_lr)
      self.cur_step += 1


class LRRangeTest(Callback):
  def __init__(
      self,
      start_lr=.00001,
      end_lr=1,
      step=None,
      steps=None,
      log_freq=None,
      plot_freq=None,
      divergence_multiplier=4
  ):
    super(LRRangeTest, self).__init__()
    self.checkpoint_path = None
    self.start_lr = start_lr
    self.end_lr = end_lr

    if not step and not steps:
      raise ValueError('step or steps must be provided')

    delta = end_lr - start_lr
    self.steps = steps or (delta / step)
    self.step = step or (delta / steps)

    self.min_loss = None
    self.avg_loss = None

    self.lrs = []
    self.losses = []

    self.plot_freq = plot_freq
    self.log_freq = log_freq

    self.divergence_multiplier = divergence_multiplier

  def __repr__(self):
    return (
      f"LRRangeTest (\n"
      f"  min_lr: {self.start_lr}\n"
      f"  max_lr: {self.end_lr}\n"
      f"  step: {self.step}\n"
      f"  steps: {self.steps}\n"
      ")"
    )

  def on_train_start(self, trainer=None):
    self.checkpoint_path = trainer.checkpoint(name='lr-range-test')

    self.lrs = [self.start_lr]
    set_param(trainer.optimizer, 'lr', self.lrs[-1])

    if self.log_freq:
      print(self)

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    self.losses.append(loss.item())
    self.avg_loss = exp_moving_avg(self.losses[-1], self.avg_loss,
                                   steps=len(self.losses))

    if self.log_freq and len(self.lrs) % self.log_freq == 0:
      print(f"lr: {self.lrs[-1]:.5f}  loss: {self.avg_loss:.5f}")

    if self.plot_freq and len(self.lrs) % self.plot_freq == 0:
      self.plot()

    if self.min_loss is None:
      self.min_loss = self.avg_loss
    elif (self.avg_loss > self.divergence_multiplier * self.min_loss) and \
        len(self.lrs) > 50:
      trainer.stop()
      return
    elif self.avg_loss < self.min_loss:
      self.min_loss = self.avg_loss

    if len(self.lrs) >= self.steps:
      trainer.stop()
      return

    self.lrs.append(self.lrs[-1] + self.step)
    set_param(trainer.optimizer, 'lr', self.lrs[-1])

  def restore(self, trainer):
    trainer.load_checkpoint(self.checkpoint_path)
    os.remove(self.checkpoint_path)

  def on_train_end(self, trainer=None):
    self.restore(trainer)

  def on_error(self, error, trainer=None):
    self.restore(trainer)

  def plot(self, **kwargs):
    plot_line(x=self.lrs, y=self.losses, xlabel='learning rate', ylabel='loss',
              **kwargs)


class CyclicalLR(Callback):
  def __init__(self, start_lr, end_lr, steps=None, step=None):
    self.start_lr = start_lr
    self.end_lr = end_lr
    self.steps = steps

    self.step = (self.end_lr - self.start_lr) / self.steps

    self.cycle_len = 2 * self.steps
    self.cur_step = 0
    self.cur_lr = self.start_lr

  def on_train_start(self, trainer=None):
    set_param(trainer.optimizer, 'lr', self.cur_lr)

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    if self.cur_step % self.cycle_len // self.steps:
      self.cur_lr -= self.step
    else:
      self.cur_lr += self.step

    if self.cur_step % 100 == 0:
      print(f'lr: {self.cur_lr:.5f}')

    set_param(trainer.optimizer, 'lr', self.cur_lr)
    self.cur_step += 1



from torch import nn

from ..data import Classes


class Classifier(nn.Module):
  classes: Classes

  def __del__(self, cls):
    raise NotImplementedError()

  def add_class(self, name, weight=None, bias=None):
    # TODO: copy other class, average multiple classes
    raise NotImplementedError()

  def remove_class(self, name):
    raise NotImplementedError()



class LinearClassifier(Classifier):
  def __init__(
      self,
      in_features,
      classes,
      bias=True,
      activation=None,
      test_activation=None
  ):
    super(Classifier, self).__init__()

    self.linear = nn.Linear(in_features, len(classes), bias=bias)
    self.classes = classes

    self.activation = activation
    self.test_activation = test_activation or activation


import numpy as np
import torch


def pad(tensor, shape, value=0):
  if tensor.shape == shape: return tensor

  if isinstance(tensor, np.ndarray):
    padded = np.zeros(shape, dtype=tensor.dtype)
  else:
    padded = torch.zeros(shape)
  if value != 0:
    padded[:] = value

  padded[tuple(map(slice, tensor.shape))] = tensor
  return padded


def pad_to_largest(tensors, value=0):
  shape = tuple(max(dim) for dim in zip(*(t.shape for t in tensors)))
  return [pad(t, shape, value) for t in tensors]


import urllib
import urllib.request
import json

from ..callbacks.base import Callback

def post(url, data):
  data = json.dumps(data).encode('utf8')
  req = urllib.request.Request(
    url, data=data, headers={'content-type': 'application/json'})
  return urllib.request.urlopen(req)


DEFAULT_CHANNEL = '#training'

def send(text, attachments=None, channel=None, username=None, icon=None,
         url=None):
  """https://api.slack.com/docs/message-attachments"""
  return post(url, {
    'text': text,
    'channel': channel or DEFAULT_CHANNEL,
    'username': username,
    'icon_emoji': icon and (icon if icon.startswith(':') else f':{icon}:'),
    'attachments': attachments,
  })


def atch(title=None, text=None, fields=None, color=None, **kwargs):
  return dict(
    title=title,
    text=text,
    fields=[
      *(fields or {}),
      *({'title': k, 'value': v} for k, v in kwargs.items())],
    color=color)




class Slack(Callback):
  # TODO: subclass Logging callback instead
  def __init__(self, channel=None, username=None, url=None, validation=False):
    self.channel = channel
    self.username = username
    self.url = url

    self.validation = validation

  def send(self, *args, **kwargs):
    send(
      *args,
      channel=self.channel,
      username=self.username,
      url=self.url,
      **kwargs
    )

  def on_train_start(self, trainer=None):
    self.send(
      text='Starting train run',
      attachments=[
        atch(experiment=trainer.name, text=trainer.description),
        atch('Configuration', f"```{trainer}```", color='good'),
      ]
    )

  def on_validation_end(self, targets=None, outputs=None, loss=None,
                        trainer=None):
    if self.validation:
      self.send(
        text=f'Completed epoch {trainer.num_epochs} with loss: {loss.item()}'
      )

  def on_error(self, error, trainer=None):
    self.send(
      text='Training run failed due to an exception',
      attachments=[
        atch(experiment=trainer.name),
        atch(epoch=trainer.num_epochs + 1),
        atch('Exception', f"```{str(error)}```", color='danger'),
      ]
    )

  def on_train_end(self, trainer=None):
    self.send(text='Train run completed')


class VectorIndex:
  @property
  def files(self):
    return []

  def save(self):
    pass

  def load(self):
    pass

  def build(self):
    pass

  def search(self):
    pass

  def search_index(self):
    pass

  def add(self, vector):
    pass

  def add_bulk(self, vectors):
    pass

  def set_bulk(self, indices, vectors):
    pass



class lazy:
  __slots__ = 'method', 'name'

  def __init__(self, method):
    self.method = method
    self.name = method.__name__

  def __get__(self, obj, cls):
    if obj:
      val = self.method(obj)
      setattr(obj, self.name, val)
      return val



from .classes import Classes
from .datasets import TransformDataset
from .loaders import TransformLoader
from .transform import Transformer


def get_dataset_name(x):
  while hasattr(x, 'dataset'):
    x = x.dataset

  if hasattr(x, 'name'):
    return x.name

  return x.__class__.__name__



import time
from matplotlib import pylab as plt

from .base import Callback


class Timing(Callback):
  def __init__(self):
    super().__init__()
    self.starts = []
    self.ends = []
    self.start_time = None

  def on_train_start(self, *args, **kwargs):
    self.start_time = time.time()

  def on_batch_start(self, *args, **kwargs):
    self.starts.append(time.time())

  def on_batch_end(self, *args, **kwargs):
    self.ends.append(time.time())

  @property
  def times(self):
    return [e - s for (s, e) in zip(self.starts, self.ends)]

  @property
  def waits(self):
    return [s - e for (s, e) in zip(self.starts, [self.start_time, *self.ends])]

  def plot(self, start=0, end=None, scatter=False):
    end = end or len(self.starts)

    fig = plt.figure(figsize=(12, 4))
    plt.grid()
    if scatter:
      plt.scatter(range(start, end), list(self.times)[start:end], label='step')
      plt.scatter(range(start, end), list(self.waits)[start:end], label='prep')
    else:
      plt.bar(range(start, end), list(self.waits)[start:end], label='prep')
      plt.bar(range(start, end), list(self.times)[start:end],
              bottom=list(self.waits)[start:end], label='step')

    plt.xlabel('step')
    plt.ylabel('seconds')
    plt.title('Train Run Timings')
    plt.legend()
    plt.show()


import pytest
import torch.cuda
from torch import nn
from torch.optim import SGD

from yann.callbacks import (
  History, HistoryPlotter, HistoryWriter, Logger, Checkpoint
)
from yann.data.datasets import TinyDigits
from yann.modules import Flatten
from yann.train import Trainer

from torchvision.models.resnet import resnet18

devices = ['cpu', 'cuda'] if torch.cuda.is_available() else ['cpu']


@pytest.mark.slow
@pytest.mark.parametrize('device', devices)
def test_train(tmpdir, device):
  """Sanity check train run"""

  model = nn.Sequential(
    nn.Conv2d(1, 20, 3),
    nn.ReLU(inplace=True),
    nn.Conv2d(20, 20, 3),
    nn.ReLU(inplace=True),
    Flatten(),
    nn.Linear(320, 10)
  )

  train = Trainer(
    root=tmpdir,
    model=model,
    dataset=TinyDigits(),
    device=device,
    optimizer=SGD(model.parameters(), lr=.01, momentum=0.9, weight_decay=.001),
    loss=nn.CrossEntropyLoss(),
    callbacks=[
      History(),
      HistoryPlotter(save=True),
      HistoryWriter(),
      Logger(batch_freq=20),
      Checkpoint()
    ]
  )

  train(10)

  assert train.checkpoints_root.is_dir()
  assert train.history.metrics

  export_path = train.export()

  assert export_path
  assert export_path.is_dir()


@pytest.mark.slow
@pytest.mark.parametrize('device', devices)
def test_train_resolved(tmpdir, device):
  from yann.data.transform import ImageTransformer

  train = Trainer(
    root=tmpdir,
    model='densenet121',
    dataset='CIFAR10',
    loss='CrossEntropy',
    optimizer='SGD',
    transform=ImageTransformer(resize=224)
  )

  train(1)



from torch.nn import AdaptiveAvgPool2d
from ..data.containers import Outputs, Inputs
from ..models import Model


class PretrainedModel(Model):
  def __init__(self, *args, **kwargs):
    self.activation = None
    super().__init__(*args, **kwargs)

  def predict(self, inputs):
    embeddings = self.features(*inputs)
    logits = self.logits(embeddings)
    activations = self.activation(logits) if self.activation else None
    return Outputs(
      embeddings=embeddings,
      logits=logits,
      activations=activations
    )


class PretrainedModelWrapper(Model):
  def __init__(self, model, activation=None, pool=None):
    super().__init__()
    self.model = model
    self.activation = activation

    self.pool = pool or AdaptiveAvgPool2d(1)
    if self.pool:
      self.model.avg_pool = pool

    if hasattr(self.model, 'avgpool') and not hasattr(self.model, 'avg_pool'):
      # the pretrained model api is inconsistent and a few cases have avgpool
      # instead of avg_pool
      self.model.avg_pool = self.model.avgpool

  def predict(self, inputs):
    if isinstance(inputs, Inputs):
      features = self.model.features(*inputs.args, **inputs.kwargs)
    else:
      features = self.model.features(inputs)

    print(features.shape)

    # pretrained models return feature maps before pooling and reshaping
    if self.pool:
      embeddings = (self.pool(features)).view(features.shape[0], -1)
    else:
      embeddings = (self.model.avg_pool(features)).view(features.shape[0], -1)

    logits = self.model.logits(features)
    activations = self.activation(logits) if self.activation else None

    return Outputs(
      embeddings=embeddings,
      logits=logits,
      activations=activations
    )


def spec_from_settings():
  pass


def spec_from_model(model):
  pass



from yann.data.io.download import Downloader


class RemoteDataset:
  def __init__(self, downloader: Downloader):
    self.downloader = downloader

  def prefetch(self, indices=None):
    for n in indices:
      self.downloader.enqueue(self[n][0])

  def __getitem__(self, item):
    return 'id', ['labels']

def test_downloader():
  downloader = Downloader('.')

  local_path = downloader['gs://foo']

  dataset = RemoteDataset(downloader)

  for uri, labels in dataset:
    downloader.enqueue(uri)


  for key, result in downloader.results:
    pass

  for key, result in downloader.as_completed():
    pass


  for key, error in downloader.errors:
    pass


  downloader.enqueue('asd')


  downloader.on_error(lambda key, error: print(error))
  downloader.on_success(lambda key, result: print('completed', key))

  downloader.complete()


  def unpack(x):
    pass


  dataset_downloader = Downloader('./')

  dataset_downloader.on_success(unpack)
  dataset_downloader.enqueue('s3://1.tar')


  dataset_downloader.get_all(['1', '2', '3'])
  dataset_downloader.stream(['1', '2', '3'])


import sys
from sklearn.metrics import classification_report, accuracy_score
from .base import Callback

from ..metrics import get_preds


class MulticlassEval(Callback):
  def __init__(self, dest=sys.stdout):
    self.dest = dest

  def __call__(self, targets=None, outputs=None, trainer=None, **kwargs):
    if trainer and trainer.dataset and hasattr(trainer.dataset, 'classes'):
      classes = trainer.dataset.classes
    else:
      classes = None

    preds = get_preds(outputs)

    preds, targets = preds.to('cpu').numpy(), targets.to('cpu').numpy()

    print(classification_report(targets, preds, target_names=classes),
          file=self.dest)

    print('Accuracy: ', accuracy_score(targets, preds))

  def on_validation_end(self, targets=None, outputs=None, loss=None,
                        trainer=None, **kwargs):
    self(targets=targets, outputs=outputs, trainer=trainer)


class MultilabelEval(Callback):
  def __call__(self, targets=None, outputs=None, trainer=None, **kwargs):
    pass



import subprocess

from ..callbacks.base import Callback
from ..utils.bash import run



def gcloud(command):
  return run(['gcloud', command])


def gsutil(command):
  return run(['gsutil', command])


def args(*flags, hyphenate=True, **kwargs):
  return ' \ \n'.join((
    *(f'--{str(n).replace("_", "-") if hyphenate else n}' for n in flags),
    *(x for x in (
      f'--{str(k).replace("_", "-") if hyphenate else k}={v}'
      if not (v is True or v is False)
      else (f'--{str(k).replace("_", "-") if hyphenate else k}' if v else '')
      for k, v in kwargs.items() if v is not None)
      if x)
  ))


def start_instance(
    name,
    zone=None,
    preemptible=True):
  command = (
    f"""gcloud compute instances create {name} \
          {
    args(
      zone=zone,
      preemptible=preemptible,
      maintenance_policy='foo',
    )
    }

    """
  )

  return run(command)


def start_dl_instance(name, ):
  pass


def copy_directory(src, dst):
  pass


def stop_instance():
  pass


def kill_instance():
  pass


def shutdown():
  return run('sudo shutdown -h now')

def gcp_sync(src, dst, exclude=None):
  if exclude:
    return subprocess.call([
      'gsutil',
      '-m',
      'rsync',
      '-r',
      '-x',
      exclude,
      src,
      dst
    ])
  else:
    return subprocess.call([
      'gsutil',
      '-m',
      'rsync',
      '-r',
      src,
      dst
    ])


class SyncCallback(Callback):
  def upload(self):
    pass

  def download(self):
    pass


class GCPSync(SyncCallback):
  pass



from collections import Counter
from inspect import isclass

from torch import nn

from ..utils import camel_to_snake


class Stack(nn.Module):
  """

  NOTE: this depends on dict ordering, which is only guaranteed for python >=3.6
  """

  def __init__(self, *layers, **named_layers):
    super().__init__()

    layer_counts = Counter()
    for l in layers:
      if l:
        name = camel_to_snake(l.__class__.__name__)
        num = layer_counts[name]
        self.add_module(name + str(num), l)
        layer_counts[name] += 1
    for name, l in named_layers.items():
      if l:
        self.add_module(name, l)

  def forward(self, x):
    for l in self.children():
      x = l(x)
    return x

  def trace(self, x):
    for l in self.children():
      x = l(x)
      yield l, x

  def __len__(self):
    return len(self._modules)

  def __getitem__(self, x):
    if isinstance(x, slice):
      layers = list(self.children())

      start, stop = x.start, x.stop
      if isinstance(x.start, nn.Module):
        start = layers.index(x.start)
      if isinstance(x.stop, nn.Module):
        stop = layers.index(x.stop)

      return self.__class__(
        **dict(list(self.named_children())[start:stop:x.step])
      )
    elif isclass(x):
      return [m for m in self.children() if isinstance(m, x)]
    else:
      return list(self.children())[x]

  def __setitem__(self, x, value):
    key = list(self._modules.keys())[x]
    setattr(self, key, value)



from torch import nn
from torch.nn import init


def kaiming(model: nn.Module):
  for module in model.modules():
    if isinstance(module, nn.Conv2d):
      init.kaiming_normal(module.weight, mode='fan_out')
      if module.bias is not None:
        init.constant(module.bias, 0)

    elif isinstance(module, nn.BatchNorm2d):
      init.constant(module.weight, 1)
      init.constant(module.bias, 0)

    elif isinstance(module, nn.Linear):
      init.normal(module.weight, std=1e-3)
      if module.bias is not None:
        init.constant(module.bias, 0)


msr = kaiming



from torch.utils.data.sampler import Sampler
import random
from collections import defaultdict


class BalancedTargetSampler(Sampler):
  def __init__(self, dataset, targets=None, size=None):
    super().__init__(dataset)
    self.dataset = dataset
    self.size = size

    self.target_to_indices = defaultdict(list)

    if not targets and hasattr(dataset, 'targets'):
      targets = dataset.targets

    if targets:
      for n, ts in enumerate(targets):
        for t in ts:
          self.target_to_indices[t].append(n)
    else:
      for n, (x, ts) in enumerate(dataset):
        for t in ts:
          self.target_to_indices[t].append(n)

    self.targets = list(self.target_to_indices)

  def __iter__(self):
    for n in range(len(self)):
      t = random.choice(self.targets)
      yield random.choice(self.target_to_indices[t])

  def __len__(self):
    return self.size or len(self.dataset)


from collections import defaultdict, OrderedDict

from functools import partial
from itertools import chain


def dedupe(items):
  return OrderedDict.fromkeys(items).keys()


def noop(x, *args, **kwargs):
  return x


def pass_args(x, *args, **kwargs):
  return x(*args, **kwargs)


class default:
  init = pass_args

  @staticmethod
  def get_names(x):
    return x.__name__,


class RegistryError(Exception):
  pass


class ResolutionError(RegistryError):
  pass


class Record:
  __slots__ = ('x', 'init')

  def __init__(self, x, init=None):
    """
    Args:
      x: registered object or class
      init: optional function used to initialize the object
    """
    self.x = x
    self.init = init


class Resolver:
  __slots__ = ('registry',)

  def __init__(self, registry: 'Registry'):
    self.registry = registry

  def resolve(
      self,
      x,
      required=False,
      validate=None,
      instance=True,
      types=None,
      init=None,
      args=None,
      kwargs=None
  ):
    initial = x
    if isinstance(x, str):
      record = self.registry[x]
      x = record.x
      if record.init:
        x = record.init(x, *(args or ()), **(kwargs or {}))

    if instance and isinstance(x, type):
      init = init or default.init
      x = init(x, *(args or ()), **(kwargs or {}))

    if not required and x is None:
      return x
    elif required and x is None:
      raise ResolutionError("Could not resolve to a value and was required")

    if types and not isinstance(x, types):
      raise ResolutionError(
        f"Failed to resolve {initial} to one of "
        f"{' '.join(str(t) for t in types)}, "
        f"got {x} instead of type {type(x)}"
      )

    if validate and not validate(x):
      raise ResolutionError(f"Couldn't validate {x}")

    return x

  def __call__(self, x, required=False, validate=None,
               instance=True, types=None, args=None, kwargs=None, init=None):
    return self.resolve(
      x,
      required=required,
      validate=validate,
      instance=instance,
      types=types,
      args=args,
      kwargs=kwargs,
      init=init
    )

  def __getattr__(self, name):
    registry = getattr(self.registry, name)
    return registry.resolve


class Registry:
  def __init__(self, types=None):
    self._records = {}
    self._children = defaultdict(Registry)
    self.resolve = Resolver(self)
    self.types = types

  def register(self, x, name=None, init=None):
    if isinstance(x, str) and name is None and init is None:
      # assuming usage as decorator, like:
      # @register("ReLu")
      # def relu():
      #   ...
      return partial(self.register, name=x)

    if self.types and not (
        issubclass(x, self.types) if isinstance(x, type)
        else isinstance(x, self.types)):
      raise RegistryError(
        f"Can't register an object of type {type(x)} in "
        f"typed registry which expects one of {self.types}")

    r = Record(x, init=init)

    if name is None:
      name = default.get_names(x)
    if isinstance(name, str):
      name = (name,)

    for n in name:
      self._records[n] = r

    return x

  def __contains__(self, name):
    if name in self._records:
      return True
    for c in self._children.values():
      if name in c:
        return True
    return False

  def has(self, x):
    """
    Checks if this registry or one of its children contain the value x
    """
    return x in self.values()

  def __call__(self, x, name=None, init=None):
    return self.register(x, name, init)

  def __setitem__(self, name, x):
    if isinstance(x, Record):
      self._records[name] = x
    else:
      self.register(x=x, name=name)

  def __getattr__(self, name: str) -> 'Registry':
    if name in self.__dict__:
      return self.__dict__[name]
    return self._children[name]

  def __setattr__(self, key, value):
    if isinstance(value, Registry):
      self._children[key] = value
    else:
      self.__dict__[key] = value

  def __getitem__(self, item) -> Record:
    if item in self._records:
      return self._records[item]

    for c in self._children.values():
      try:
        return c[item]
      except KeyError:
        pass

    raise KeyError(
      f"Couldn't find key: '{item}', "
      f"valid options include: {', '.join(self._records.keys())}")

  def values(self):
    return dedupe((
      *(r.x for r in self._records.values()),
      *chain(*(c.values() for c in self._children.values()))
    ))

  def items(self):
    return (
      *((k, r.x) for k, r in self._records.items()),
      *chain(*(c.items() for c in self._children.values()))
    )

  def __len__(self):
    return len(self.values())

  def index(
      self,
      module,
      types=None,
      get_names=None,
      include=None,
      init=None
  ):
    """
    Indexes a module. If types are specified will only include entries of
    given type.
    Args:
      module:
      types:
      get_names:
      include:

    Returns:

    """
    for item in module.__dict__.values():
      if isinstance(item, (int, str)):
        continue
      if types:
        if not (
            isinstance(item, types) or
            (isinstance(item, type)) and issubclass(item, types)):
          continue

      if include and not include(item):
        continue

      names = get_names(item) if get_names else default.get_names(item)

      self.register(item, name=names, init=init)

  def register_subclasses(self, cls: type, init=None):
    self.register(cls, init=init)
    for scls in cls.__subclasses__():
      self.register(scls, init=init)

  def update(self, items, init=None):
    for x in items:
      self.register(x, init=init)

  def __str__(self):
    parts = [
      ''.join(f"  '{name}': {r.x.__name__}  ({r.init})\n"
              for name, r in self._records.items())
    ]

    for name, c in self._children.items():
      parts.append(f"\n{name}:\n")
      parts.append(str(c))

    return ''.join(parts)



from torch import nn

from ..data.containers import Outputs


class ModelMixin:
  def __init__(self, *args, **kwargs):
    # Need to store this for when we save the model,
    # so that we can restore it later without knowing them
    self._init_args = {
      'args': args,
      'kwargs': kwargs
    }
    super().__init__(*args, **kwargs)

  def predict(self, inputs) -> Outputs:
    """
    Inference mode predict call
    """
    return None

  @classmethod
  def load(cls, *args, **kwargs) -> 'Model':
    return cls(*args, **kwargs)


class Model(ModelMixin, nn.Module):
  pass



from .base import Callback


class Checkpoint(Callback):
  def __init__(self, freq=1, load_latest=False):
    self.freq = freq
    self.load_latest = load_latest

  def on_train_start(self, trainer=None):
    if self.load_latest:
      # trainer.load_checkpoint()
      pass

  def on_epoch_end(self, epoch, loss=None, metrics=None, trainer=None):
    if epoch % self.freq == 0 :
      trainer.checkpoint()




import logging

import torch

from .registry import Registry, pass_args

## Configure Registry

registry = Registry()

# Datasets
import torchvision.datasets

from torch.utils.data import Dataset

registry.dataset.index(
  torchvision.datasets,
  types=(Dataset,),
  init=lambda D, root=None, download=True, **kwargs: \
    D(root=root or f'~/.torch/datasets/{D.__name__}',
      download=download)
)

from ..data.datasets import imagenette
registry.dataset.index(
  imagenette,
  types=(Dataset,),
  init=lambda D, root=None, download=True, **kwargs: \
    D(root=root or f'~/.torch/datasets/',
      download=download)
)


# Losses
registry.loss.index(
  torch.nn.modules.loss,
  types=(torch.nn.modules.loss._Loss,),
  get_names=lambda x:
  (x.__name__, x.__name__[:-len('Loss')]) if x.__name__.endswith('Loss') else (
    x.__name__,)
)

import torch.nn.functional as F

registry.loss.index(
  F,
  include=lambda x: hasattr(x, '__name__') and 'loss' in x.__name__.lower(),
  get_names=lambda x: (x.__name__, x.__name__[:-len('_loss')]) if
  x.__name__.endswith('_loss') else (x.__name__,)
)

from ..modules import loss

registry.loss.update((
  F.cross_entropy,
  F.binary_cross_entropy,
  F.binary_cross_entropy_with_logits,
  loss.soft_target_cross_entropy,
  loss.SoftTargetCrossEntropyLoss
))

# Optimizers
from torch.optim import optimizer

registry.optimizer.index(
  torch.optim,
  types=(optimizer.Optimizer,)
)

registry.optimizer['SGD'].init = \
  lambda SGD, params, lr=.01, momentum=.9, weight_decay=1e-4: \
    SGD(params, lr=lr, momentum=momentum, weight_decay=weight_decay)

# LR Schedulers
from torch.optim import lr_scheduler

registry.lr_scheduler.index(
  lr_scheduler,
  types=(lr_scheduler._LRScheduler,)
)

# Models
from torchvision import models


def is_public_callable(x):
  return (
      hasattr(x, '__name__')
      and not x.__name__.startswith('_')
      and callable(x)
  )


registry.model.torchvision.index(
  models,
  init=pass_args,
  include=is_public_callable
)

try:
  import pretrainedmodels.models
except ImportError:
  logging.debug("Couldn't register pretrainedmodels models because it's not "
                "installed")
else:
  registry.model.pretrainedmodels.index(
    pretrainedmodels.models,
    init=pass_args,
    include=is_public_callable
  )

from .. import metrics
registry.metric.update((
  metrics.accuracy,
  metrics.average_precision,
  metrics.average_precision_at_k,
  metrics.coverage_error,
))

registry.metric.register(
  metrics.top_3_accuracy,
  name='top_3_accuracy'
)

registry.metric.register(
  metrics.top_5_accuracy,
  name='top_5_accuracy'
)

registry.metric.register(
  metrics.top_10_accuracy,
  name='top_10_accuracy'
)


from contextlib import contextmanager

__version__ = '0.0.29'

import torch
from torch import nn
from .config.setup import registry

register = registry
resolve = registry.resolve

from .utils import to_numpy

default_device = torch.device('cuda') \
  if torch.cuda.is_available() else torch.device('cpu')


def seed(val=1, deterministic=False):
  import numpy as np
  import torch
  import random
  random.seed(val)
  np.random.seed(val)
  torch.manual_seed(val)
  try:
    torch.cuda.manual_seed(val)

    if deterministic:
      torch.cuda.deterministic = True
  except:
    pass
  return val


def benchmark():
  from torch.backends import cudnn
  cudnn.benchmark = True


def detect_anomalies(val=True):
  import torch.autograd
  torch.autograd.set_detect_anomaly(val)


def evaluate(model, batches, device=None):
  for x, y in batches:
    if device:
      x, y = x.to(device), y.to(device)

    model.eval()
    with torch.no_grad():
      pred = model(x)

    yield x, y, pred


def predict_multicrop(model, inputs, reduce='mean'):
  batch_size, num_crops, *sample_shape = inputs.shape
  flat_preds = model(inputs.view(-1, *sample_shape))
  outputs = flat_preds.view(batch_size, num_crops, -1)
  if reduce:
    outputs = getattr(torch, reduce)(outputs, 1)
  return outputs


class Multicrop(torch.nn.Module):
  def __init__(self, model):
    super().__init__()
    self.model = model

  def forward(self, inputs, *rest):
    return predict_multicrop(self.model, inputs)


def set_param(x, param, val):
  for group in x.param_groups:
    group[param] = val


def scale_param(x, param, mult):
  for group in x.param_groups:
    group[param] *= mult


def trainable(parameters):
  return (p for p in parameters if p.requires_grad)


# TODO: handle batchnorm
def freeze(parameters):
  for p in parameters:
    p.requires_grad = False


def unfreeze(parameters):
  for p in parameters:
    p.requires_grad = True


def replace_linear(model, num_outputs, layer_name='last_linear'):
  setattr(
    model,
    layer_name,
    nn.Linear(
      getattr(model, layer_name).in_features,
      num_outputs)
  )


def to_fp16(model):
  # https://discuss.pytorch.org/t/training-with-half-precision/11815
  # https://github.com/csarofeen/examples/tree/fp16_examples_cuDNN-ATen/imagenet
  # https://github.com/NVIDIA/apex
  # https://github.com/NVIDIA/apex/tree/master/apex/amp
  model.half()
  for layer in model.modules():
    if isinstance(layer, torch.nn.BatchNorm2d):
      layer.float()


class HyperParams:
  def __init__(self, **args):
    self.__dict__.update(args)

  def __setattr__(self, key, value):
    raise AttributeError('Updating properties is not permitted')

  def __getitem__(self, item):
    if isinstance(item, (tuple, list)):
      return tuple(self.__dict__[k] for k in item)
    return self.__dict__[item]

  def fork(self, **args):
    return HyperParams(**{**self.__dict__, **args})

  def __repr__(self):
    return (
      'HyperParams('
      f"{', '.join(f'{k}={v}' for k, v in self.__dict__.items())}"
      ')'
    )

  def __str__(self):
    return (
        'HyperParams (\n' +
        ''.join('  {}={}\n'.format(k, v) for k, v in self.__dict__.items()) +
        ')'
    )

  def __len__(self):
    return len(self.__dict__)

  def __hash__(self):
    return hash(tuple(sorted(self.items())))

  def keys(self):
    return self.__dict__.keys()

  def values(self):
    return self.__dict__.values()

  def items(self):
    return self.__dict__.items()

  def grid(self, **args):
    raise NotImplementedError()

  def inject(self, scope=None, uppercase=True):
    scope = globals() if scope is None else scope
    for k, v in self.items():
      scope[k.upper() if uppercase else k] = v

  @classmethod
  def collect(cls, scope=None, types=(int, str, float, bool),
              upper_only=True, lowercase=True):
    scope = globals() if scope is None else scope

    d = {}
    for k, v in scope.items():
      if types and not isinstance(v, types):
        continue
      if upper_only and not k.isupper():
        continue

      d[k.lower() if lowercase else k] = v

    return cls(**d)

  def __eq__(self, other):
    return hash(self) == hash(other)


@contextmanager
def evalmode(*modules, grad=False):
  if grad:
    training = (m.training for m in modules)
    try:
      for m in modules: m.eval()
      yield
    finally:
      for m, train in zip(modules, training):
        if train:
          m.train()
  else:
    with torch.no_grad():
      training = (m.training for m in modules)
      try:
        for m in modules: m.eval()
        yield
      finally:
        for m, train in zip(modules, training):
          if train:
            m.train()



import logging

import datetime
import pathlib
import torch
import torch.nn
from torch.utils.data import DataLoader
from typing import Optional

from .. import callbacks as yann_callbacks
from .. import resolve, evaluate
from ..data import TransformDataset, get_dataset_name, Classes
from ..export import export
from ..inference.predict import Classifier
from ..train.base import BaseTrainer
from ..utils.decorators import lazy
from ..utils import counter


def timestr(d=None):
  return f"{(d or datetime.datetime.utcnow()).strftime('%y-%m-%dT%H:%M:%S')}"


def get_model_name(model):
  if isinstance(model, torch.nn.DataParallel):
    model = model.module

  if hasattr(model, 'name'):
    return model.name

  return model.__class__.__name__


class Trainer(BaseTrainer):
  model: torch.nn.Module
  lr_scheduler: Optional[torch.optim.lr_scheduler._LRScheduler]
  loader: DataLoader

  def __init__(
      self,
      model,
      dataset=None,
      optimizer=None,
      loss=None,
      loader=None,
      sampler=None,
      num_workers=8,
      transform=None,
      lr_scheduler=None,
      lr_batch_step=False,
      callbacks=None,
      device=None,
      trainable_parameters=None,
      batch_size=16,
      val_dataset=None,
      val_loader=None,
      val_transform=None,
      classes=None,
      parallel=False,
      name=None,
      description=None,
      root='./train-runs/',
      metrics=None,
      collate=None
  ):
    super().__init__()

    self.model = resolve.model(
      model,
      required=True,
      validate=callable
    )
    if parallel:
      self.model = torch.nn.DataParallel(self.model)

    self.loss = resolve.loss(
      loss,
      required=True,
      validate=callable,
    )
    self.optimizer = resolve.optimizer(
      optimizer,
      args=(trainable_parameters or self.model.parameters(),),
      required=True,
      validate=lambda x: hasattr(x, 'step')
    )

    self.dataset = resolve.dataset(
      dataset,
      required=not loader,
    )

    if classes:
      self.classes = (
        classes if isinstance(classes, Classes)
        else Classes(classes)
      )
    elif hasattr(self.dataset, 'classes') and isinstance(self.dataset.classes, Classes):
      self.classes = self.dataset.classes
    else:
      self.classes = None

    if transform:
      self.dataset = TransformDataset(self.dataset, transform)
    self.transform = transform

    self.batch_size = batch_size

    self.loader = loader or DataLoader(
      self.dataset,
      batch_size=self.batch_size,
      pin_memory=True,
      shuffle=False if sampler else True,
      sampler=sampler,
      num_workers=num_workers,
      **({'collate_fn': collate} if collate else {})
    )

    self.val_dataset = resolve.dataset(
      val_dataset,
      required=False,
    )

    self.val_transform = val_transform or transform
    if self.val_transform:
      self.val_dataset = TransformDataset(self.val_dataset, self.val_transform)

    self.val_loader = val_loader or (val_dataset and DataLoader(
      self.val_dataset,
      batch_size=batch_size,
      shuffle=False,
      pin_memory=True,
      num_workers=num_workers
    ))

    self.lr_scheduler = resolve.lr_scheduler(
      lr_scheduler,
      kwargs=dict(optimizer=self.optimizer)
    )
    self.lr_batch_step = lr_batch_step

    self.num_samples = 0
    self.num_steps = 0
    self.num_epochs = 0

    self.time_created = datetime.datetime.utcnow()

    self.name = name or (
      f"{get_dataset_name(self.loader)}-{get_model_name(self.model)}"
    )
    self.description = description

    self.root = pathlib.Path(root) / self.name / timestr(self.time_created)
    self.root.mkdir(parents=True, exist_ok=True)

    self.history = None
    has_history = False
    if callbacks:
      for c in callbacks:
        if isinstance(c, yann_callbacks.History):
          self.history = c
          has_history = True
          break
    self.history = self.history or yann_callbacks.History(*(metrics or ()))

    self.callbacks = callbacks or [
      yann_callbacks.Logger()
    ]

    if not has_history:
      self.callbacks.append(self.history)

    # make sure history callback is called first
    self.callbacks = sorted(
      self.callbacks,
      key=lambda x: 0 if x is self.history else 1)

    self.function_callback = None

    if device:
      self.device = torch.device(device) if isinstance(device, str) else device
      self.to(self.device)
    else:
      self.device = None

  def __setattr__(self, key, value):
    if key == 'optimizer':
      if hasattr(self, 'lr_scheduler') and hasattr(self.lr_scheduler, 'optimizer'):
        self.lr_scheduler.optimizer = value
    if key == 'loader':
      if hasattr(self, 'dataset') and hasattr(value, 'dataset'):
        super(Trainer, self).__setattr__('dataset', value.dataset)
      if hasattr(value, 'batch_size'):
        super(Trainer, self).__setattr__('batch_size', value.batch_size)
    if key == 'batch_size' and hasattr(self, 'batch_size') and self.batch_size != key:
      if hasattr(self, 'loader') and self.loader:
        raise ValueError(
          'Cannot modify batch_size because a loader is defined '
          'and modifying batch size of a loader is not supported, '
          'try creating and setting a new loader instead'
        )
      if key == 'dataset' and hasattr(self, 'dataset') and self.dataset \
          and hasattr(self, 'loader') and self.loader:
        raise ValueError(
          'Cannot modify dataset because a loader is defined '
          'and modifying dataset of a loader is not supported, '
          'try creating and setting a new loader instead'
        )

    logging.debug(f"setting '{key}' to {value}")
    super(Trainer, self).__setattr__(key, value)


  def to(self, device=None):
    self.device = device

    if self.model:
      self.model.to(self.device)

  def on(self, event, callback=None):
    if not self.function_callback:
      self.function_callback = yann_callbacks.FunctionCallback()
      self.callbacks.append(self.function_callback)

    if callback:
      self.function_callback.on(event, callback)
      return self
    else:
      def decorated(func):
        self.function_callback.on(event, func)
        return func

      return decorated

  def step(self, inputs, target):
    self.model.train()
    self.optimizer.zero_grad()

    outputs = self.model(inputs)
    loss = self.loss(outputs, target)
    loss.backward()
    self.optimizer.step()

    self.num_steps += 1
    self.num_samples += len(inputs)

    return outputs, loss

  def batches(self):
    for inputs, targets in self.loader:
      if self.device:
        yield inputs.to(self.device), targets.to(self.device)
      else:
        yield inputs, targets

  def validate(self, loader=None, device=None):
    loader = loader or self.val_loader
    device = device or self.device

    self.model.eval()

    self.on_validation_start()

    ts, os = [], []
    with torch.no_grad():
      for inputs, targets, outputs in evaluate(self.model, loader, device):
        self.on_validation_batch(inputs=inputs, targets=targets,
                                 outputs=outputs)
        ts.append(targets)
        os.append(outputs)

      ts = torch.cat(ts)
      os = torch.cat(os)
      loss = self.loss(os, ts)
    self.on_validation_end(targets=ts, outputs=os, loss=loss)
    return loss

  def run(self, epochs=None):
    self._stop = False
    try:
      self.on_train_start()

      for _ in counter(end=epochs):
        self.on_epoch_start(epoch=self.num_epochs)
        for batch_idx, (inputs, targets) in enumerate(self.batches()):
          self.on_batch_start(
            batch=batch_idx,
            inputs=inputs,
            targets=targets
          )
          try:
            outputs, loss = self.step(inputs, targets)
          except KeyboardInterrupt as e:
            self.stop()
            break
          except Exception as e:
            self.on_batch_error(batch=batch_idx, error=e)
            if self._stop: break
            raise e

          self.on_batch_end(
            batch=batch_idx,
            inputs=inputs,
            targets=targets,
            outputs=outputs,
            loss=loss
          )

          if self.lr_scheduler and self.lr_batch_step:
            self.lr_scheduler.step(epoch=self.num_steps)

          if self._stop: break
        if self._stop: break

        if self.val_loader:
          val_loss = self.validate()
          if self.lr_scheduler and not self.lr_batch_step:
            if 'metrics' in self.lr_scheduler.step.__code__.co_varnames:
             self.lr_scheduler.step(metrics=val_loss, epoch=self.num_epochs)
            else:
              self.lr_scheduler.step(epoch=self.num_epochs)
        elif self.lr_scheduler and not self.lr_batch_step:
          self.lr_scheduler.step(epoch=self.num_epochs)

        self.on_epoch_end(epoch=self.num_epochs)
        self.num_epochs += 1

      self.on_train_end()
    except Exception as e:
      self.on_error(e)
      raise e

  def __call__(self, *args, **kwargs):
    self.run(*args, **kwargs)

  def stop(self):
    self._stop = True

  @lazy
  def predict(self):
    return Classifier(
      model=self.model,
      classes=self.classes,
      preprocess=self.transform,
      postprocess=None
    )

  def get_checkpoint_name(self):
    return (
      f"{self.name}-steps_{self.num_steps}.th"
    )

  @property
  def checkpoints_root(self):
    root = self.root / 'checkpoints'
    root.mkdir(parents=True, exist_ok=True)
    return root

  def checkpoint(self, root='./', name=None):
    state = self.state_dict()
    path = str(self.checkpoints_root /
               (f"{name}.th" if name else
                f"{timestr()}-epoch-{self.num_epochs:03d}-"
                f"steps-{self.num_steps:05d}.th"))
    torch.save(state, path)
    return path

  def load_checkpoint(self, path, metadata=True):
    data = torch.load(path)
    self.load_state_dict(data, metadata=metadata)

  def export(self, path=None, trace=False, meta=None, postprocess=None):
    path = path or self.root / 'exports' / timestr()
    export(
      model=self.model,
      preprocess=self.val_transform,
      postprocess=postprocess,
      classes=self.classes,
      trace=trace and next(iter(self.loader)),
      path=path,
      meta=dict(
        name=self.name,
        root=str(self.root),
        dataset=get_dataset_name(self.dataset),
        num_steps=self.num_steps,
        **(meta or {})
      )
    )

    return path

  def state_dict(self):
    data = {
      'metadata': {
        'num_steps': self.num_steps,
        'num_samples': self.num_samples,
        'num_epochs': self.num_epochs,
        'batch_size': self.batch_size,
        'name': self.name,
        'time_created': datetime.datetime.utcnow()
      }
    }

    for k, v in self.__dict__.items():
      if hasattr(v, 'state_dict'):
        data[k] = {
          'state_dict': v.state_dict(),
          'class': v.__class__.__name__
        }

    return data

  def load_state_dict(self, data, metadata=True):
    """
    TODO: add a way to specify which parts should be loaded (ex: model only)
    """
    skipped = set()
    for k, v in data.items():
      if 'state_dict' in v and hasattr(self, k):
        getattr(self, k).load_state_dict(v['state_dict'])
      else:
        skipped.add(k)

    if metadata and 'metadata' in data:
      skipped.discard('metadata')
      for k, v in data['metadata'].items():
        setattr(self, k, v)

    if skipped:
      logging.warning(f'skipped {skipped} when loading checkpoint')

  def __str__(self):
    return f"""
name: {self.name}
root: {self.root}
batch_size: {self.batch_size}
device: {self.device}

MODEL
=====

{self.model}


DATASET
=======

{self.loader.dataset}


LOADER
======

{self.loader}

LOSS
====

{self.loss}


OPTIMIZER
=========

{self.optimizer}

SCHEDULER
=========

{self.lr_scheduler}


PROGRESS
========
epochs: {self.num_epochs}
steps: {self.num_steps}
samples: {self.num_samples}
"""



from collections import OrderedDict, defaultdict
from contextlib import contextmanager

from datetime import datetime
import torch.cuda

from ..viz.plot import plot_timeline


class Task:
  __slots__ = ('name', 'start_time', 'end_time', 'meta', 'sync')

  def __init__(self, name=None, start=None, end=None, meta=None, sync=False):
    self.name = name
    self.start_time = start
    self.end_time = end
    self.meta = meta or {}
    self.sync = sync

  def start(self, time=None, meta=None, sync=None):
    if meta:
      self.meta.update(meta)
    sync = sync if sync is not None else self.sync
    if sync and torch.cuda.is_available():
      torch.cuda.synchronize()

    self.start_time = time or datetime.now()

  def end(self, time=None, meta=None, sync=None):
    sync = sync if sync is not None else self.sync
    if sync and torch.cuda.is_available():
      torch.cuda.synchronize()

    self.end_time = time or datetime.now()

    if meta:
      self.meta.update(meta)

  @classmethod
  def begin(cls, name=None, meta=None, sync=None):
    t = cls(name=name, meta=meta, sync=sync)
    t.start()
    return t

  @property
  def seconds(self):
    if self.start_time is None or self.end_time is None:
      return None
    return (self.end_time - self.start_time).total_seconds()

  def __enter__(self):
    self.start()
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.end()


class Timer:
  def __init__(self, name=None, log=False):
    self.tasks = []
    self.name = name
    self.log = log

    self.active_tasks = {}

  def start(self, name, sync=True, **meta):
    task = self.task(name, sync=sync, **meta)
    if self.log: print('Started', name)

    if task in self.active_tasks:
      raise ValueError(f'Nesting tasks is not allowed, "{name}" was already started and not finished')
    self.active_tasks[name] = task

  def end(self, name, sync=True, **meta):
    task = self.active_tasks.pop(name)
    if not task:
      raise ValueError(f"{name} is not an active task so can't be ended")
    task.end(sync=sync, meta=meta)

    if self.log:
      print('Ended', task.name, ', took', task.seconds, 'seconds')

  def task(self, name, sync=True, **meta):
    task = Task.begin(name=name, meta=meta, sync=sync)
    self.tasks.append(task)
    return task

  def __enter__(self):
    self.start(self.name)
    return self

  def __exit__(self, exc_type, exc_val, exc_tb):
    self.end(self.name)

  def plot(self):
    plot_timeline(self.tasks)


import os

from . import set_param
from .train import train, Trainer
from torch.optim import lr_scheduler

def lr_range_test(trainer: Trainer,  min_lr=.00001, max_lr=1, steps=None,
                  step=None, log_freq=None, restore=True):
  # assert max_lr > min_lr
  if restore:
    checkpoint_path = trainer.checkpoint(name='lr-range-test')
  else:
    checkpoint_path = None

  steps = steps or len(trainer.loader)
  step = step or ((max_lr - min_lr) / steps)

  set_param(trainer.optimizer, 'lr', min_lr)
  try:
    cur_lr = min_lr
    cur_step = 0

    while cur_step < steps:
      for x, y, pred, loss in train(trainer.model, trainer.loader,
                                    trainer.optimizer, trainer.loss,
                                    trainer.device):
        yield (cur_lr, loss)

        if log_freq and cur_step % log_freq == 0:
          print(cur_lr, loss.item())

        cur_lr += step
        cur_step += 1
        set_param(trainer.optimizer, 'lr', cur_lr)



  finally:
    if restore:
      print('loading checkpoint')
      trainer.load_checkpoint(checkpoint_path)
      os.remove(checkpoint_path)



class SchedulerSequence(lr_scheduler._LRScheduler):
  def __init__(self, schedulers):
    self.schedulers = schedulers
    super(SchedulerSequence, self).__init__()


class Warmup(lr_scheduler._LRScheduler):
  def __init__(self, scheduler: lr_scheduler._LRScheduler, ):
    pass



def callback(method):
  def wrapped_method(self, *args, **kwargs):
    ret = method(self, *args, **kwargs)
    for c in self.callbacks:
      if hasattr(c, method.__name__):
        getattr(c, method.__name__)(*args, **kwargs, trainer=self)
    return ret

  return wrapped_method


class BaseTrainer:
  def __init__(self):
    self._stop = False

  def run(self, *args, **kwargs):
    pass

  def batches(self, *args, **kwargs):
    return []

  def step(self, *args, **kwargs):
    pass

  def __call__(self, *args, **kwargs):
    self.run(*args, **kwargs)

  def stop(self):
    self._stop = True

  def checkpoint(self):
    pass

  @classmethod
  def from_checkpoint(cls, path):
    pass

  @callback
  def on_train_start(self):
    pass

  @callback
  def on_epoch_start(self, epoch):
    pass

  @callback
  def on_batch_start(self, batch, inputs, targets):
    pass

  @callback
  def on_batch_end(self, batch, inputs, targets, outputs, loss):
    pass

  @callback
  def on_batch_error(self, batch, error):
    pass

  @callback
  def on_error(self, error):
    pass

  @callback
  def on_epoch_end(self, epoch):
    pass

  @callback
  def on_validation_start(self):
    pass

  @callback
  def on_validation_batch(self, inputs, targets, outputs):
    pass

  @callback
  def on_validation_end(self, targets=None, outputs=None, loss=None):
    pass

  @callback
  def on_train_end(self):
    pass











from torch.nn import Module


class Reshape(Module):
  method = None

  def __init__(self, *dims):
    super(Reshape, self).__init__()
    self.dims = dims

  def forward(self, input):
    return getattr(input, self.method)(*self.dims)

  def state_dict(self, destination=None, prefix='', keep_vars=False):
    return {
      'dims': self.dims
    }

  def load_state_dict(self, state_dict, strict=True):
    self.dims = state_dict['dims']


class Squeeze(Reshape):
  method = 'squeeze'


class Permute(Reshape):
  method = 'permute'


class Transpose(Reshape):
  method = 'transpose'


class View(Reshape):
  method = 'view'


class Flatten(Reshape):
  def forward(self, input, *rest):
    return input.view(input.size(0), -1)


def flatten(input):
  return input.view(input.size(0), -1)


class FlattenSequences(Module):
  def forward(self, input, *rest):
    return flatten_sequence(input)


def flatten_sequence(seq_batch):
  seq_len, batch_size, *rest = seq_batch.size()
  return seq_batch.view(seq_len * batch_size, -1)


class Infer(Module):
  def __init__(self, cls, *args, **kwargs):
    super(Infer, self).__init__()
    self.cls = cls

    self.args = args
    self.kwargs = kwargs

    self.module = None

  @property
  def was_inferred(self):
    return self.module is not None

  def forward(self, x):
    if self.module is None:
      self.module = self.cls(x.shape[1], *self.args, **self.kwargs)
    return self.module(x)

  @classmethod
  def shed(cls, module):
    pass


import pytest
from torch.optim import SGD
from torch.utils.data import Dataset
from torchvision import datasets
from torchvision.datasets import MNIST, FashionMNIST

from yann.config.registry import Registry, RegistryError


def test_registry():
  r = Registry()

  assert len(r) == 0, "Registry didn't start out empty"
  r.register(MNIST, name='mnist')
  assert len(r) == 1

  assert 'mnist' in r._records
  assert r._records['mnist'].x is MNIST

  r2 = Registry(types=(Dataset,))

  r2.register(MNIST)

  with pytest.raises(RegistryError):
    r2.register(SGD)


def test_registry_nesting():
  r = Registry()

  assert isinstance(r.dataset, Registry)

  r.dataset.register(MNIST)
  r.dataset.fashion.register(FashionMNIST)

  assert len(r.dataset) == 2
  assert len(r.dataset.fashion) == 1


def test():
  registry = Registry()
  yann = lambda: None
  yann.registry = registry
  yann.register = registry
  yann.resolve = registry.resolve

  yann.register(MNIST, 'mnist')

  assert yann.registry.resolve('mnist', instance=False) is MNIST

  yann.register.dataset(MNIST)

  ds = list(yann.registry.dataset.values())

  yann.register.optimizers(
    SGD,
    init=lambda SGD, *args, parameters=None, **kwargs: SGD(parameters, )
  )


def test_indexing():
  registry = Registry()
  yann = lambda: None
  yann.registry = registry
  yann.register = registry
  yann.resolve = registry.resolve

  yann.registry.dataset.index(datasets, types=(Dataset,))
  assert yann.resolve.dataset('MNIST', instance=False) is datasets.MNIST

  assert yann.registry.has(datasets.MNIST)
  assert 'MNIST' in yann.registry
  assert 'MNIST' in yann.registry.dataset
  assert 'MNIST' not in yann.registry.loss

  class CustomDataset(Dataset):
    def __init__(self, required_arg):
      self.required_arg = required_arg

  yann.register.dataset(CustomDataset)

  dset = CustomDataset('value')
  assert yann.resolve.dataset(dset, instance=True) is dset

  @yann.register.dataset
  class Foobar(Dataset):
    pass

  assert yann.resolve.dataset('Foobar', instance=False) is Foobar

  @yann.register('ReLU')
  def relu(x):
    return max(0, x)

  assert yann.resolve('ReLU') is relu
  assert yann.resolve('ReLU')(2) == 2
  assert yann.resolve('ReLU')(-1) == 0

  assert yann.resolve(relu)(-1) == 0


def test_yann_registry():
  import yann

  assert len(yann.registry.dataset)
  assert len(yann.registry.loss)
  assert len(yann.registry.optimizer)

  yann.resolve('MNIST', required=True)



from .classification import Trainer
from ..callbacks import Callback


def step(model, inputs, targets, optimizer, loss, callback: Callback = None):
  model.train()
  optimizer.zero_grad()

  pred = model(inputs)
  loss = loss(pred, targets)

  loss.backward()
  optimizer.step()

  return inputs, targets, pred, loss


def train(model, batches, optimizer, loss, device=None, step=step, callback: Callback=None):
  for inputs, targets in batches:
    if device:
      inputs, targets = inputs.to(device), targets.to(device)
    yield step(model, inputs, targets, optimizer, loss)


import pyarrow


pyarrow.serialization.register_torch_serialization_handlers(
  pyarrow.serialization.default_serialization_context()
)


def deserialize_arrow(buf):
  return pyarrow.deserialize(buf)


def serialize_arrow(x):
  return pyarrow.serialize(x).to_buffer()


serialize = serialize_arrow
deserialize = deserialize_arrow


def to_bytes(string: str) -> bytes:
  return string.encode(encoding='utf-8', errors='strict')


def to_unicode(b: bytes) -> str:
  return b.decode(encoding='utf-8', errors='strict')



from collections import defaultdict
import operator
from functools import reduce


class InvertedIndex:
  def __init__(self, items):
    self._index = defaultdict(set)
    self.update(items)

  def update(self, items):
    for key, vals in items:
      for v in vals:
        self._index[v].add(key)

  def __getitem__(self, vals):
    if isinstance(vals, tuple):
      return set(reduce(operator.and_, (self._index[v] for v in vals)))
    else:
      return self._index[vals]

  def get(self, vals, not_vals=None):
    if not_vals:
      return self[vals] - self[not_vals]
    else:
      return self[vals]


import json
import os
import pickle as pkl
import tarfile
from collections import namedtuple
import csv
import gzip


def save_pickle(obj, path, mode='wb'):
  with open(str(path), mode) as f:
    pkl.dump(obj, f)


def load_pickle(path, mode='rb'):
  with open(str(path), mode) as f:
    return pkl.load(f)


def save_json(obj, path, mode='w'):
  with open(str(path), mode) as f:
    json.dump(obj, f)


def load_json(path, mode='r'):
  with open(str(path), mode) as f:
    return json.load(f)


def tar_dir(path, dest=None, zip=True):
  path = str(path)
  dest = str(dest or path)

  ext, mode = ('.tar.gz', 'w:gz') if zip else ('.tar', 'w')

  if not dest.endswith(ext):
    dest = os.path.splitext(dest)[0] + ext

  with tarfile.open(dest, mode) as tar:
    tar.add(path)


def lines(path, mode='r'):
  with open(path, mode=mode) as f:
    yield from f


def write_lines(items, path, mode='w'):
  with open(path, mode=mode) as f:
    f.writelines(items)


def untar(path):
  with tarfile.open(path) as tar:
    tar.extractall()


def iter_csv(path, header=True, tuples=True, sep=',', quote='"', **kwargs):
  with open(path) as f:
    reader = csv.reader(f, delimiter=sep, quotechar=quote, **kwargs)
    if header:
      if tuples:
        reader = iter(reader)
        h = next(reader)
        Row = namedtuple('Row', h)

        for r in reader:
          yield Row(*r)
      else:
        reader = iter(reader)
        h = next(reader)
        for r in reader:
          yield dict(zip(h, r))
    else:
      yield from reader


def write_csv(data, path, header=None):
  with gzip.open(path, 'wt') if path.endswith('.gz') else open(path, 'w') as f:
    writer = csv.writer(f)
    if header:
      writer.writerow(header)
    for row in data:
      writer.writerow(row)




import sys
import logging

from .base import Callback


class Logger(Callback):
  def __init__(self, batch_freq=128, dest=sys.stdout):
    super(Logger, self).__init__()

    self.dest = dest

    self.batch_freq = batch_freq
    self.batch_string = None
    self.logged_batch_shapes = False

  def log(self, *args, sep='\t', **kwargs):
    if kwargs:
      print(
        *args,
        sep.join(f'{k}: {v}' for k, v in kwargs.items()),
        file=self.dest)
    else:
      print(*args, file=self.dest)

  def on_train_start(self, trainer=None):
    self.log('Starting training\n', trainer)

  def on_batch_end(self, batch, inputs, targets, outputs, loss, trainer=None):
    if batch % self.batch_freq == 0:
      if not self.logged_batch_shapes:
        try:
          self.log("\nBatch inputs shape:", tuple(inputs.size()),
                   "\nBatch targets shape:", tuple(targets.size()),
                   "\nBatch outputs shape:", tuple(outputs.size()), '\n')
        except:
          pass
        self.logged_batch_shapes = True

      if self.batch_string:
        self.log(self.batch_string.format(
          batch=batch, **({m: v[-1] for m, v in
                           trainer.history.metrics.items()})))
      else:
        self.log(
          batch=f'{batch:>8}',
          **({m: f'{v[-1]:.4f}' for m, v in trainer.history.metrics.items()}))

  def on_epoch_start(self, epoch, trainer=None):
    self.log('\nStarting epoch', epoch)

    self.log(f'''
OPTIMIZER
=========

{trainer.optimizer}


PROGRESS
========
epochs: {trainer.num_epochs}
steps: {trainer.num_steps}
samples: {trainer.num_samples}\n''')

  def on_validation_start(self, trainer=None):
    self.log('\nStarting Validation')

  def on_validation_end(self, loss=None, outputs=None, targets=None,
    trainer=None):
    self.log('\nCompleted Validation')
    self.log(
      epoch=trainer.num_epochs,
      steps=trainer.history.val_steps[-1],
      **{m: f'{vals[-1]:.4f}' for m, vals in trainer.history.val_metrics.items()}
    )

  def on_epoch_end(self, epoch, loss=None, metrics=None, trainer=None):
    self.log('Completed epoch', epoch,)

  def on_train_end(self, trainer=None):
    self.log('Completed training run. \n\n')





from itertools import zip_longest
import logging


class DatasetWrapper:
  def __init__(self, dataset):
    self.dataset = dataset

  def __getitem__(self, item):
    return self.dataset[item]

  def __len__(self):
    return len(self.dataset)



class IncludeIndex(DatasetWrapper):
  def __getitem__(self, idx):
    z = super().__getitem__(idx)

    if isinstance(z, (tuple, list)):
      return (*z, idx)

    return (z, idx)


class Slice(DatasetWrapper):
  def __init__(self, dataset, start=0, end=None):
    super(Slice, self).__init__(dataset)

    self.start = start
    self.end = end if end is not None else len(dataset)

  def __getitem__(self, idx):
    if idx >= len(self):
      raise IndexError(f'Index out of bounds {idx}')
    return self.dataset[self.start + idx]

  def __len__(self):
    return self.end - self.start


class Subset(DatasetWrapper):
  pass


class LookupCache(DatasetWrapper):
  """
  Cache lookups

  Default to storing cache in memory so avoid caching large datasets or large inputs

  pass a preprocess function to transform the outputs before caching (useful
  for fine tuning on precomputed embeddings)

  # TODO: add joblib.Memory wrapper as a disk based cache option
  """

  def __init__(self, dataset, preprocess=None, cache=None):
    super().__init__(dataset)

    self.cache = cache or {}
    self.preprocess = preprocess

  def __getitem__(self, item):
    if item in self.cache:
      return self.cache[item]

    x = super().__getitem__(item)
    if self.preprocess:
      x = self.preprocess(x)

    self.cache[item] = x
    return x

  def set(self, key, value):
    self.cache[key] = value

  def update(self, keys, values):
    for k, v in zip(keys, values):
      self.set(k, v)


class TransformDataset(DatasetWrapper):
  def __init__(self, dataset, transforms):
    super().__init__(dataset)
    self.transforms = transforms if isinstance(transforms, tuple) else (
      transforms,)

  def __getitem__(self, idx):
    return tuple(
      t(x) if t else x
      for (x, t) in zip_longest(self.dataset[idx], self.transforms)
    )

  def __repr__(self):
    return (
      f'{self.__class__.__name__}('
      f'\nDataset: {repr(self.dataset)}'
      f'\nTransforms: {repr(self.transforms)}'
      f'\n)')


class Noisy(DatasetWrapper):
  def __init__(self, dataset, p=0.1):
    super().__init__(dataset)
    self.p = p


class Imbalanced(DatasetWrapper):
  pass


class SemiSupervised(DatasetWrapper):
  """TODO: hide subset of labels"""
  pass


class SwallowErrors(DatasetWrapper):
  def __init__(self, dataset, errors=None):
    super(SwallowErrors, self).__init__(dataset)
    self.errors = errors or Exception

  def __getitem__(self, item):
    try:
      return self.dataset[item]
    except KeyboardInterrupt as e:
      raise e
    except self.errors as e:
      logging.warning(e)
      return None


class VariableLength(DatasetWrapper):
  def __init__(self, dataset, max_size=1e20):
    super(VariableLength, self).__init__(dataset)
    self.max_size = max_size

  def __len__(self):
    return self.max_size

  def __getitem__(self, idx):
    return self.dataset[idx % len(self.dataset)]



from scipy.misc import imresize



def draw_mask(img, mask, blend=.5, cmap=None, interp='cubic'):
    if not cmap:
      import matplotlib.pylab as plt
      cmap = plt.get_cmap('jet')
    if isinstance(cmap, str):
      import matplotlib.pylab as plt
      cmap = plt.get_cmap(cmap)

    if mask.shape[:2] != img.shape[:2]:
        mask = imresize(mask, img.shape[:2], interp=interp)
    return (cmap(mask)[:,:,:3] * 255 * blend + img * (1-blend)).round().astype('uint8')


def class_activation_maps(features, weights, classes=None, normalize=True):
  num_samples, num_channels, rows, cols = features.shape

  classes = classes or list(range(weights.shape[0]))

  maps = []
  for sample in features:
    class_maps = {}
    for c in classes:
      blended_channels = (
            weights[c] @ sample.reshape(num_channels, rows * cols)).reshape(
        rows, cols)
      if normalize:
        x = blended_channels - blended_channels.min()
        x = x / x.max() * 255
        class_maps[c] = x.cpu().detach().numpy()
      else:
        class_maps[c] = blended_channels.cpu().detach().numpy()
    maps.append(class_maps)
  return maps


#
# class ClassActivationMaps:
#   def __init__(self):
#     self.conv_hook_handle = None
#     self.conv_features = None
#     self.cams = None
#
#   def hook(self, last_conv):
#     self.conv_hook_handle = last_conv.register_forward_hook()
#
#   def record_features(self, module, inputs, outputs):
#     self.conv_features = outputs
#
#   def draw(self, img):
#     pass
#
#   def show(self):
#     pass


import click


@click.group()
def cli():
  pass


@cli.command()
@click.option('-n', '--name', default=None)
@click.option('-d', '--dataset', default=None)
@click.option('-opt', '--optimizer', default='SGD', show_default=True)
@click.option('-m', '--model', default='resnet18', show_default=True)
@click.option('-e', '--epochs', default=10, show_default=True)
@click.option('-l', '--loss')
@click.option('-cp', '--checkpoint')
@click.option('-c', '--continue')
def train(
    name,
    model,
    dataset,
    loss=None,
    transform=None,
    optimizer='SGD',
    checkpoint=None,
    lr=0.01,
    momentum=.9,
    epochs=10
):
  """Train model"""
  from .train import Trainer
  from .callbacks import get_callbacks

  t = Trainer(
    name=name,
    model=model,
    optimizer=optimizer,
    dataset=dataset,
    transform=transform,
    loss=loss,
    callbacks=get_callbacks(interactive=False)
  )

  if checkpoint:
    t.load_checkpoint(checkpoint)

  t(epochs)


@cli.command()
def tune():
  pass


@cli.command()
def evaluate(self):
  """Evaluate model on data"""
  pass


@cli.command()
def validate():
  pass


@cli.command()
@click.argument('names', nargs=-1)
def resolve(names):
  import yann

  if len(names) == 1:
    print(yann.resolve(names[0]))
  else:
    x = getattr(yann.resolve, names[0])
    for n in names[1:-1]:
      x = getattr(x, n)
    print(x(names[-1]))


@cli.command()
@click.argument('names', nargs=-1)
def registry(names):
  """List contents of registry"""
  import yann

  if not names:
    print(yann.registry)
  else:
    x = yann.registry
    for n in names:
      x = getattr(x, n)
    print(x)


def main():
  cli()


if __name__ == '__main__':
  cli()



class styles(dict):
  def __init__(self, *args, **props):
    super().__init__()
    if args:
      if isinstance(args[0], dict):
        self.update(args[0])
    if isinstance(args[0], str):
      for l in args[0].split(';'):
        if l:
          k, v = l.split(':')

          self[k.replace('-', '_').strip()] = v.strip()

    self.update(props)

  def __str__(self):
    return ' '.join(f"{k.replace('_', '-')}: {v};"
                    for k, v in self.items())


class Node:
  NAME = None
  EMPTY = False

  def __init__(self, *children, style=None, **props):
    self.children = children
    self.props = props
    self.style = style if isinstance(style, styles) else styles(style)

  @property
  def name(self):
    return self.NAME or self.__class__.__name__

  def __str__(self):
    if self.EMPTY:
      return f"""
         <{self.name} {' '.join(
        f'{k}="{v}"' for k, v in self.props.items())}  style="{self.style}"/>
      """

    return f"""
    <{self.name} {' '.join(
      f'{k}="{v}"' for k, v in self.props.items())} style="{self.style}">
       {' '.join(str(c) for c in self.children)}
    </{self.name}>
    """

  __repr__ = __str__

  def render(self, target='notebook'):
    if target == 'notebook':
      from IPython.core.display import display, HTML
      return display(HTML(str(self)))

    return str(self)


class div(Node): pass


class span(Node): pass


class img(Node):
  EMPTY = True


class p(Node): pass


class h1(Node): pass



def zero_out_nans(module, grad_input, grad_output):
  """
  Backwards hook to turn NaNs in gradients to 0
  """
  for grad in grad_input:
    grad[grad != grad] = 0  # technically shouldn't modify inputs



from yann.data.containers import Container, Inputs


class TestContainer:
  def test_set_and_get_item(self):
    c = Container(1, 2, 3, x=5, y=6)

    assert list(c) == [1, 2, 3, 5, 6]
    assert c[-1] == 6
    assert c[:2] == [1, 2]

    assert c.args == (1, 2, 3)
    assert c.kwargs == {'x': 5, 'y': 6}

    c[3] = 9
    assert c.x == 9

    input = Inputs(image='sfsd', mask=[0, 1, 2, 3])

    image, *rest = input

    assert image == input.image == 'sfsd'



import torch
from pretrainedmodels.models import squeezenet1_1
from torch import nn

import yann
from yann.contrib.pretrained import PretrainedModelWrapper
from yann.utils.timer import Timer


def test_pretrained():
  timer = Timer(log=True)

  with timer.task('initialize model'):
    model = squeezenet1_1(1000, pretrained=None)

  wrapped = PretrainedModelWrapper(model, activation=nn.Softmax())

  inputs = torch.rand(1, 3, 255, 255)

  with yann.evalmode(wrapped), timer.task('model predict'):
    outputs = wrapped.predict(inputs)
  with yann.evalmode(model), timer.task('model predict'):
    model_outputs = model(inputs)

  assert torch.all(outputs.logits == model_outputs)
  assert outputs.logits.shape
  assert outputs.embeddings.shape
  assert outputs.activations.shape



from collections import defaultdict

class Callback:
  def on_train_start(self, trainer=None):
    pass

  def on_epoch_start(self, epoch=None, trainer=None):
    pass

  def on_batch_start(self, batch=None, inputs=None, targets=None, trainer=None):
    pass

  def on_batch_end(self, batch=None, inputs=None, targets=None, outputs=None, loss=None,
                   trainer=None):
    pass

  def on_epoch_end(self, epoch=None, loss=None, metrics=None, trainer=None):
    pass

  def on_validation_start(self, trainer=None):
    pass

  def on_validation_batch(self, inputs=None, targets=None, outputs=None, trainer=None):
    pass

  def on_validation_end(self, targets=None, outputs=None, loss=None,
                        trainer=None):
    pass

  def on_train_end(self, trainer=None):
    pass

  def on_batch_error(self, batch=None, error=None, trainer=None):
    pass

  def on_error(self, error=None, trainer=None):
    pass


class FunctionCallback(Callback):
  _valid_names = {
    'train_start',
    'epoch_start',
    'batch_start',
    'batch_end',
    'epoch_end',
    'validation_start',
    'validation_batch',
    'validation_end',
    'train_end',
    'error',
  }

  def __init__(self, callbacks=None):
    self.callbacks = callbacks or defaultdict(list)

  def on(self, event, callback):
    if event not in self._valid_names:
      raise ValueError(
        f'{event} is not a valid option, '
        f'must be one of {self._valid_names}')
    callbacks = self.callbacks[event]
    if callback not in callbacks:
      callbacks.append(callback)

  def on_train_start(self, *args, **kwargs):
    for f in self.callbacks['train_start']:
      f(*args, **kwargs)

  def on_epoch_start(self, *args, **kwargs):
    for f in self.callbacks['epoch_start']:
      f(*args, **kwargs)

  def on_batch_start(self, *args, **kwargs):
    for f in self.callbacks['batch_start']:
      f(*args, **kwargs)

  def on_batch_end(self, *args, **kwargs):
    for f in self.callbacks['batch_end']:
      f(*args, **kwargs)

  def on_epoch_end(self, *args, **kwargs):
    for f in self.callbacks['epoch_end']:
      f(*args, **kwargs)

  def on_validation_start(self, *args, **kwargs):
    for f in self.callbacks['validation_start']:
      f(*args, **kwargs)

  def on_validation_batch(self, *args, **kwargs):
    for f in self.callbacks['validation_batch']:
      f(*args, **kwargs)

  def on_validation_end(self, *args, **kwargs):
    for f in self.callbacks['validation_end']:
      f(*args, **kwargs)

  def on_train_end(self, *args, **kwargs):
    for f in self.callbacks['train_end']:
      f(*args, **kwargs)

  def on_error(self, *args, **kwargs):
    for f in self.callbacks['error']:
      f(*args, **kwargs)


class TempCallback(Callback):
  def __init__(self, callback, steps=None, epochs=None):
    self.callback = callback
    self.steps = steps
    self.epochs = epochs

  def unregister(self):
    raise NotImplementedError()

  def on_batch_end(self, batches, *args, **kwargs):
    if self.steps and self.steps < batches:
      self.unregister()

  def on_epoch_end(self, epoch=None, loss=None, metrics=None, trainer=None):
    if self.epochs and self.epochs < epoch:
      self.unregister()


from torch import nn
from ...modules import Stack, Flatten
from .base import RecognitionModel


class VGG(RecognitionModel):
  Conv = nn.Conv2d
  Norm = nn.BatchNorm2d
  Downsample = nn.MaxPool2d
  Activation = nn.ReLU

  Reduce = nn.AvgPool2d

  in_channels = 3

  channels = [
    (64,),
    (128,),
    (256,),
    (512,),
    (512,)
  ]

  def __init__(self, num_classes):
    super(VGG, self).__init__()
    self.num_classes = num_classes
    self.backbone = self.build_backbone()

    self.embeddings = Stack(
      pool=self.Reduce(kernel_size=1, stride=1),
      flatten=Flatten()
    )

    self.classifier = nn.Linear(self.channels[-1][-1], self.num_classes)

  def forward(self, inputs):
    features = self.backbone(inputs)
    embeddings = self.embeddings(features)
    logits = self.classifier(embeddings)
    return logits

  def build_backbone(self):
    layers = []

    prev_channels = self.in_channels
    for channels in self.channels:
      for c in channels:
        layers.append(
          Stack(
            conv=self.Conv(prev_channels, c, kernel_size=3, padding=1),
            norm=self.Norm(c),
            activation=self.Activation(inplace=True)
          )
        )
        prev_channels = c
      layers.append(self.Downsample(kernel_size=2, stride=2))
    return Stack(*layers)


class VGG11(VGG):
  channels = [
    (64,),
    (128,),
    (256,) * 2,
    (512,) * 2,
    (512,) * 2,
  ]


class VGG13(VGG):
  channels = [
    (64,) * 2,
    (128,) * 2,
    (256,) * 2,
    (512,) * 2,
    (512,) * 2,
  ]


class VGG16(VGG):
  channels = [
    (64,) * 2,
    (128,) * 2,
    (256,) * 3,
    (512,) * 3,
    (512,) * 3,
  ]


class VGG19(VGG):
  channels = [
    (64,) * 2,
    (128,) * 2,
    (256,) * 4,
    (512,) * 4,
    (512,) * 4,
  ]



from io import BytesIO
from PIL import Image
from itertools import zip_longest
import base64


def base64str(img):
  s = BytesIO()
  img.save(s, format='JPEG')
  return (
      "data:image/jpeg;base64,"
      + base64.b64encode(s.getvalue()).decode("utf-8")
  )


def show_images(paths, labels=None, urls=False, w=400, h=400):
  from IPython.core.display import display, HTML

  tags = []
  for x, l in zip_longest(paths, labels or []):
    if urls:
      src = x
    else:
      img = Image.open(x) if isinstance(x, str) else x
      img.thumbnail((w, h))
      src = base64str(img)

    if l:
      tags.append(
        f'''
        <div style="display: inline-block; padding: 3px">
          <img 
            style="max-width: {w}px; max-height: {h}px; margin: 3px;"
             src={src} />
          <p>{l if isinstance(l, str) else ', '.join(l)}</p>
        </div>
        '''
      )
    else:
      tags.append(
        f'''
        <img 
          style="max-width: {w}px; max-height: {h}px; margin: 3px;"
           src={src} />
        '''
      )
  return display(HTML(f'<div>{"".join(tags)}</div>'))


import numpy as np


class TargetTransformer:
  def encode(self, x, many=True):
    pass

  def decode(self, x, many=True):
    pass

  def state_dict(self):
    pass

  def load_state_dict(self, data):
    pass


class Classes(TargetTransformer):
  valid_encodings = {
    'index',
    'one_hot',
    'normalized_one_hot'
  }

  def __init__(
      self,
      names,
      meta=None,
      default_encoding='index'
  ):
    self.names = list(names)
    self.indices = {c: i for i, c in enumerate(self.names)}
    self.meta = meta
    
    self.counts = None

    self.dtype = 'float32'

    assert default_encoding in self.valid_encodings, \
      f'default_encoding must be one of {self.valid_encodings}, got {default_encoding}'
    self.default_encoding = default_encoding
    
  @classmethod
  def ordered(cls, num, meta=None, default_encoding='index'):
    return Classes(range(num), meta=meta, default_encoding=default_encoding)

  def __repr__(self):
    c = min(len(self.names) // 2, 3)

    return (
      f"Classes(\n" 
      f"  count={len(self)},\n" 
      f"  default_encoding={self.default_encoding}\n"
      f"  names=[{', '.join(self.names[:c])}, ..., {', '.join(self.names[-c:])}]\n"
      # f"  encoded={self.encode(self.names[:c])}, ..., {self.encode(self.names[-c:])}\n"
      f")"
    )

  def state_dict(self):
    return {
      'names': self.names,
      'meta': self.meta,
      'default_encoding': self.default_encoding
    }

  def load_state_dict(self, data):
    if 'classes' in data:
      # fr backwards compatibility since classes was renamed to names
      self.names = data['classses']
    else:
      self.names = data['names']
    self.indices = {c: i for i, c in enumerate(self.names)}
    self.meta = data['meta']
    self.default_encoding = data['default_encoding']

  def __getitem__(self, idx):
    return self.names[idx]

  def __iter__(self):
    return iter(self.names)

  def __contains__(self, cls):
    return cls in self.indices

  def __len__(self):
    return len(self.names)

  def __eq__(self, other):
    return self.indices == other.indices

  def encode(self, seq, encoding=None):
    return getattr(self, (encoding or self.default_encoding) + '_encode')(seq)

  def decode(self, encoded, encoding=None):
    return getattr(self, (encoding or self.default_encoding) + '_decode')(
      encoded)

  def index_encode(self, classes):
    if isinstance(classes, (str, int)):
      return self.indices[classes]
    return [self.indices[c] for c in classes]

  def index_decode(self, indices):
    if isinstance(indices, int):
      return self[indices]
    return [self[idx] for idx in indices]

  def one_hot_encode(self, classes):
    if isinstance(classes, (str, int)):
      classes = [classes]
    y = np.zeros(len(self), dtype=self.dtype)
    y[[self.indices[c] for c in classes]] = 1
    return y

  def normalized_one_hot_encode(self, classes):
    if isinstance(classes, (str, int)):
      classes = [classes]
    y = np.zeros(len(self), dtype=self.dtype)
    y[[self.indices[c] for c in classes]] = 1
    y = np.true_divide(y, y.sum())
    return y

  def ranked_decode(self, scores):
    indices = np.argsort(scores)
    return [(self.names[i], scores[i]) for i in indices][::-1]


def smooth(y, eps=.1, num_classes=None):
  if not num_classes:
    if len(y.shape) == 1:
      num_classes = len(y)
    else:
      num_classes = y.shape[1]
  return y * (1 - eps) + eps * (1.0 / num_classes)



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
  long_description = fh.read()

setup(
  name='yann',
  version='0.0.29',
  description='yet another neural network library',
  long_description=long_description,
  long_description_content_type="text/markdown",
  url='https://github.com/michalwols/yann',
  author='Michal Wolski',
  author_email='michal@bite.ai',
  license='MIT',
  packages=find_packages(),
  entry_points={
    'console_scripts': ['yann=yann.cli:main'],
  },
  extras_require={
    'cli': ['click>=6.7'],
    'pretrainedmodels': ['pretrainedmodels']
  },
  install_requires=[
    'numpy',
    'scipy',
    'scikit-learn',
    'torch>=1.0.0',
    'matplotlib',
    'torchvision'
  ],
  python_requires='>=3.6',
  zip_safe=False)



import numpy as np
import re
import torch
from PIL import Image
import sys



def camel_to_snake(text):
  s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
  return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


def truthy(items):
  return [x for x in items if x]


class Obj(dict):
  __setattr__ = dict.__setitem__
  __getattr__ = dict.__getitem__


def almost_equal(t1, t2, prec=1e-12):
  return torch.all(torch.lt(torch.abs(torch.add(t1, -t2)), prec))


def equal(t1, t2):
  return torch.all(t1 == t2)


def randimg(*shape, dtype='uint8'):
  return Image.fromarray((np.random.rand(*shape) * 255).astype(dtype))


def progress(it, num=None):
  if not num:
    try:
      num = len(it)
    except:
      num = None

  if num:
    for n, x in enumerate(it, 1):
      sys.stdout.write(f"\r{n} / {num}")
      sys.stdout.flush()
      yield x
  else:
    for n, x in enumerate(it, 1):
      sys.stdout.write(f"\r{n}")
      sys.stdout.flush()
      yield x


def repeat(val):
  while True:
    yield val


def counter(start=0, end=None, step=1):
  current = start
  while end is None or (end and (current < end)):
    yield current
    current += step


def to_numpy(x):
  if isinstance(x, np.ndarray):
    return x
  if torch.is_tensor(x):
    return x.to('cpu').numpy()
  return np.array(x)



import shutil

import torch
from torch import nn
from torchvision import transforms

from yann.export import export, load
from yann.utils import almost_equal, equal, randimg


class Net(nn.Module):
  def __init__(self, num_classes=10, input_shape=(1, 32, 32)):
    super().__init__()
    self.num_classes = num_classes
    self.input_shape = input_shape

    self.conv = nn.Sequential(
      nn.Conv2d(1, 3, 3),
      nn.MaxPool2d(2),
      nn.ReLU(inplace=True),
      nn.Conv2d(3, 3, 3),
      nn.MaxPool2d(2),
      nn.ReLU(inplace=True),
    )

  def forward(self, input, *args, **kwargs):
    return self.conv(input)


def check_loaded_model(loaded, model):
  pass


def test_export_traced(tmpdir):
  NUM_CLASSES = 10
  model = Net(NUM_CLASSES)

  preprocess = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.4,), (0.4,))
  ])

  path = tmpdir

  if path.exists():
    shutil.rmtree(str(path))

  classes = [str(n) for n in range(NUM_CLASSES)]

  export(
    path=path,
    model=model,
    trace=torch.rand(1, 1, 32, 32),
    preprocess=preprocess,
    classes=classes

  )

  expected_files = [
    'model.traced.th',
    'preprocess.pkl',
    'classes.json',
    'requirements.txt',
    'env.yml'
  ]

  for name in expected_files:
    assert (path / name).exists()

  loaded = load(path)

  assert loaded.model
  assert loaded.preprocess
  assert loaded.classes
  assert loaded.classes == classes

  input = torch.rand(1, 1, 32, 32)

  loaded.model.eval()
  model.eval()
  assert almost_equal(model(input), loaded.model(input), 1e-12)

  img = randimg(32, 32, 3)
  assert equal(preprocess(img), loaded.preprocess(img))

  assert classes == loaded.classes


def test_export_state_dict(tmpdir):
  NUM_CLASSES = 10
  model = Net(NUM_CLASSES)

  preprocess = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.4,), (0.4,))
  ])

  path = tmpdir

  if path.exists():
    shutil.rmtree(str(path))

  export(
    path=path,
    model=model,
    state_dict=True,
    preprocess=preprocess,
    classes=[str(n) for n in range(NUM_CLASSES)]
  )

  expected_files = [
    'model.state_dict.th',
    'preprocess.pkl',
    'classes.json',
  ]

  for name in expected_files:
    assert (path / name).exists()

  loaded = load(path)


def test_export_pickled(tmpdir):
  NUM_CLASSES = 10
  model = Net(NUM_CLASSES)

  preprocess = transforms.Compose([
    transforms.ToTensor(),
    transforms.Normalize((0.4,), (0.4,))
  ])

  path = tmpdir

  if path.exists():
    shutil.rmtree(str(path))

  export(
    path=path,
    model=model,
    state_dict=False,
    preprocess=preprocess,
    classes=[str(n) for n in range(NUM_CLASSES)]

  )

  expected_files = [
    'model.th',
    'preprocess.pkl',
    'classes.json',
  ]

  for name in expected_files:
    assert (path / name).exists()

  loaded = load(path)
